' Copyright 2009 Sven Strickroth <email@cs-ware.de>
'
' This file is part of OSMtracker.
'
' OSMtracker is free software: you can redistribute it and/or modify
' it under the terms of the GNU General Public License version 3 as
' published by the Free Software Foundation.
'
' OSMtracker is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
' GNU General Public License for more details.
'
' You should have received a copy of the GNU General Public License
' along with OSMtracker. If not, see <http://www.gnu.org/licenses/>.

Option Explicit On
Option Strict On

Imports System.Windows.Forms
Public Class EditButton
    Private button As Button = Nothing
    Public Sub New(ByRef button As Button)
        InitializeComponent()
        Me.button = button
        txtTitle.Text = button.Title
        cmbTag.Text = button.Tag
        If button.GotoPage > 0 Then
            cmbGotoPage.Text = button.GotoPage.ToString
        End If
        txtExecPath.Text = button.ExecPath
        txtIconFilename.Text = button.IconFileName
        txtHotkey.Text = button.HotKey
        txtSound.Text = button.SoundFileName
        ' move all GroupBoxes
        For Each cntl As Control In Me.Controls
            If cntl.GetType.Name = "GroupBox" AndAlso cntl.Name <> "gbGotoPage" Then
                cntl.Left = gbGotoPage.Left
                cntl.Height = gbGotoPage.Height
                cntl.Visible = False
            End If
        Next
        cmbTag_SelectedIndexChanged(Nothing, Nothing)
        For i As Integer = 1 To ButtonXML.HardwareButtons
            cmbHardKey.Items.Add(i.ToString)
        Next
        If button.HardKey > 0 Then
            cmbHardKey.SelectedIndex = button.HardKey
        Else
            cmbHardKey.SelectedIndex = -1
        End If
    End Sub
    Private Sub btnOK_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnOK.Click
        Me.DialogResult = System.Windows.Forms.DialogResult.OK
        button.Title = txtTitle.Text
        button.Tag = cmbTag.Text
        button.IconFileName = txtIconFilename.Text
        button.SoundFileName = txtSound.Text
        If button.HotKey <> "" Then
            ButtonXML.RemoveHotKeyMapping(button)
        End If
        If txtHotkey.Text.Length > 0 Then
            button.HotKey = Char.Parse(txtHotkey.Text)
            ButtonXML.AddHotKeyMapping(button)
        Else
            button.HotKey = Nothing
        End If
        If button.HardKey > 0 Then
            ButtonXML.RemoveHwBtnMapping(button.HardKey)
        End If
        If cmbHardKey.SelectedIndex > 0 Then
            button.HardKey = cmbHardKey.SelectedIndex
            ButtonXML.AddHwBtnMapping(button)
        Else
            button.HardKey = 0
        End If
        If cmbTag.Text = "__goto" OrElse cmbTag.Text = "_goto" Then
            Try
                button.GotoPage = CInt(cmbGotoPage.Text)
            Catch err As Exception
                button.GotoPage = -1
                MsgBox("Invalid GotoPage-number!")
                Exit Sub
            End Try
        End If
        If cmbTag.Text = "__exec" OrElse cmbTag.Text = "__loadbtnfile" Then
            button.ExecPath = txtExecPath.Text
        End If
        Me.Close()
    End Sub
    Private Sub btnCancel_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnCancel.Click
        Me.DialogResult = System.Windows.Forms.DialogResult.Cancel
        Me.Close()
    End Sub
    Private Sub cmbTag_SelectedIndexChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cmbTag.SelectedIndexChanged
        gbGotoPage.Visible = cmbTag.Text = "__goto" OrElse cmbTag.Text = "_goto"
        gbExecPath.Visible = cmbTag.Text = "__exec" OrElse cmbTag.Text = "__loadbtnfile"
    End Sub
    Private Sub txtHotkey_TextChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles txtHotkey.TextChanged
        hotkeyusedwarning.Text = ""
        If txtHotkey.Text.Length = 1 Then
            Dim chararray() As Char = txtHotkey.Text.ToCharArray
            Dim keyaction As New KeyAction(chararray(0))
            If ButtonXML.KeyActions.Contains(keyaction) Then
                keyaction = ButtonXML.KeyActions.Find(keyaction).Value
                If keyaction.btn.Tag <> cmbTag.Text Then
                    hotkeyusedwarning.Text = txtHotkey.Text + " already used for " + keyaction.btn.ToString
                End If
            End If
        End If
    End Sub
    Private Sub cmbHardKey_SelectedIndexChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cmbHardKey.SelectedIndexChanged
        hwbtnusedwarning.Text = ""
        If cmbHardKey.SelectedIndex > 0 Then
            Dim btn As Button = ButtonXML.HWButtonActions(cmbHardKey.SelectedIndex - 1)
            If btn IsNot Nothing AndAlso btn.Tag <> "" AndAlso btn.Tag <> cmbTag.Text Then
                hwbtnusedwarning.Text = cmbHardKey.Text + " already used for " + btn.ToString
            End If
        End If
    End Sub
End Class
