' Copyright 2007 Rubke <interrubke@gmail.com>
' Copyright 2008 - 2009 Sven Strickroth <email@cs-ware.de>
' Copyright 2008 - 2009 Jozef Vince <jozef.vince@freemap.sk>
'
' This file is part of OSMtracker.
'
' OSMtracker is free software: you can redistribute it and/or modify
' it under the terms of the GNU General Public License version 3 as
' published by the Free Software Foundation.
'
' OSMtracker is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
' GNU General Public License for more details.
'
' You should have received a copy of the GNU General Public License
' along with OSMtracker. If not, see <http://www.gnu.org/licenses/>.

Option Explicit On
Option Strict On

Imports System.IO
Imports System.Xml

''' <summary>
''' Represents one page in our multipage button layout definition file button.xml
''' </summary>
Public Class Page
    ''' <summary>
    ''' Page ID, should be unique
    ''' </summary>
    Public ID As Integer
    ''' <summary>
    ''' When AutoPageAdvance is enabled after clicking on button it will take you to this page number 
    ''' (except bottons with __goto command, and special button without support of AutoPageAdvance command)
    ''' </summary>
    Public GotoPage As Integer
    ''' <summary>
    ''' Page title displayed on the top of the page
    ''' </summary>
    Public Title As String
    ''' <summary>
    ''' Page layout and mode
    '''  * DefaultAutoSize - up to 12 buttons, with autosizing and auto arange to 3x4 or 3x4 - default behavior for backward compatibility
    '''  * Fixed - no auto sizing, but every button has its own position and size, virtually unlimited number of buttons with same design asi the default mode
    '''  * BackgroudImage - in this mode you have to provide one fixed size background image, where your buttons are prerendered in eyecatching form - no text and no other images are drawn, only press sensitive areas are defined
    ''' </summary>
    Public PageMode As PagePresentationMode = PagePresentationMode.DefaultAutoSize
    ''' <summary>
    ''' False if AutoSize layout has to be rearranged to reflect current ButtonPanel dimension
    ''' </summary>
    Public IsPrepared As Boolean = False

    ''' <summary>
    ''' Full path to page background image
    ''' </summary>
    Public PageImageFileName As String = ""
    ''' <summary>
    ''' Bitmap of backgroundImage
    ''' </summary>
    Public PageImage As Bitmap

    ''' <summary>
    ''' Buttons collection contained on this page
    ''' </summary>
    Public Buttons As ButtonsCollection = New ButtonsCollection

    ''' <summary>
    ''' Equalses the specified other object.
    ''' </summary>
    ''' <param name="otherObject">The other object.</param>
    ''' <returns></returns>
    Public Overrides Function Equals(ByVal otherObject As Object) As Boolean
        If Not otherObject.GetType.Equals(GetType(Page)) Then
            Equals = False
        Else
            If DirectCast(otherObject, Page).ID = ID Then
                Equals = True
            Else
                Equals = False
            End If
        End If
    End Function

    Public Function Load(ByRef oPage As Xml.XmlElement, ByVal storageFolder As String) As Boolean
        Dim appFolder As String = GetAppPath()

        If oPage.Attributes("id") IsNot Nothing Then
            ID = Integer.Parse(oPage.Attributes("id").InnerText)
        End If
        If oPage.Attributes("gotoPage") IsNot Nothing Then
            GotoPage = Integer.Parse(oPage.Attributes("gotoPage").InnerText)
        Else
            GotoPage = 1
        End If

        If oPage.Attributes("title") IsNot Nothing Then
            Title = oPage.Attributes("title").InnerText
        End If
        If oPage.Attributes("extended") IsNot Nothing Then
            Select Case Integer.Parse(oPage.Attributes("extended").InnerText)
                Case 0
                    PageMode = PagePresentationMode.DefaultAutoSize
                Case 1
                    PageMode = PagePresentationMode.Fixed
                Case 2
                    PageMode = PagePresentationMode.BackgroundImage
            End Select
        Else
            PageMode = PagePresentationMode.DefaultAutoSize
        End If
        If oPage.Attributes("pageImage") IsNot Nothing Then
            Dim pageImageFile As String = oPage.Attributes("pageImage").InnerText

            If File.Exists(storageFolder & IO.Path.DirectorySeparatorChar & "img" & IO.Path.DirectorySeparatorChar & pageImageFile) Then
                PageImageFileName = storageFolder & IO.Path.DirectorySeparatorChar & "img" & IO.Path.DirectorySeparatorChar & pageImageFile
            Else
                PageImageFileName = appFolder & IO.Path.DirectorySeparatorChar & "img" & IO.Path.DirectorySeparatorChar & pageImageFile
            End If

            If File.Exists(PageImageFileName) Then
                PageImage = New System.Drawing.Bitmap(PageImageFileName)
            Else
                ' background image is not found
                If PageMode = PagePresentationMode.BackgroundImage Then
                    ' fall back to fixed method
                    PageMode = PagePresentationMode.Fixed
                End If
            End If
        End If

        Buttons.Load(oPage.SelectSingleNode("buttons"), storageFolder)
        Return (True)
    End Function

#If ButtonEditor Then
    ''' <summary>
    ''' Saves the buttons.
    ''' </summary>
    ''' <param name="oPage">xmlelement to serialize pages</param>
    Public Function Save(ByRef xmlDoc As XmlDocument, ByRef oPage As XmlElement) As Boolean
        Dim oButtons As XmlElement

        ' store page attributes
        oPage.SetAttribute("id", ID.ToString)
        oPage.SetAttribute("title", Title)
        If GotoPage <> 1 Then
            oPage.SetAttribute("gotoPage", GotoPage.ToString)
        End If
        If PageMode <> PagePresentationMode.DefaultAutoSize Then
            oPage.SetAttribute("extended", [Enum].Format(GetType(PagePresentationMode), PageMode, "d"))
        End If
        oButtons = xmlDoc.CreateElement("buttons")
        oPage.AppendChild(oButtons)
        Buttons.Save(xmlDoc, oButtons)
    End Function

#End If

    ''' <summary>
    ''' Updates the auto size layout.
    ''' </summary>
    ''' <param name="width">The width.</param>
    ''' <param name="height">The height.</param>
    ''' <param name="topMargin">The top margin.</param>
    Public Sub UpdateAutoSizeLayout(ByVal width As Integer, ByVal height As Integer, ByVal topMargin As Integer)
        If PageMode = PagePresentationMode.DefaultAutoSize Then
            Dim ColumnWidth As Integer
            Dim RowHeight As Integer

            Dim MarginX As Integer = 10
            Dim MarginY As Integer = 10
            Dim MarginTop As Integer = topMargin + 5 - 10

            Dim w As Integer = width
            Dim h As Integer = (height - MarginTop)

            Dim Rectangle As Rectangle

            ' use Height here, because in normal mode there is h+1=w for me, but (Height>Width)=true looks better
            Dim OrientationPortrait As Boolean = (height > width)
            If OrientationPortrait Then
                ColumnWidth = CInt((w - (MarginX * 4)) / 3)
                RowHeight = CInt((h - (MarginY * 5)) / 4)
                Rectangle = New Rectangle(MarginX, (MarginY + MarginTop), ColumnWidth, RowHeight)
                PositionButton(1, Rectangle)
                Rectangle = New Rectangle(((MarginX * 2) + ColumnWidth), (MarginY + MarginTop), ColumnWidth, RowHeight)
                PositionButton(2, Rectangle)
                Rectangle = New Rectangle(((MarginX * 3) + (ColumnWidth * 2)), (MarginY + MarginTop), ColumnWidth, RowHeight)
                PositionButton(3, Rectangle)
                Rectangle = New Rectangle(MarginX, (((MarginY * 2) + RowHeight) + MarginTop), ColumnWidth, RowHeight)
                PositionButton(4, Rectangle)
                Rectangle = New Rectangle(((MarginX * 2) + ColumnWidth), (((MarginY * 2) + RowHeight) + MarginTop), ColumnWidth, RowHeight)
                PositionButton(5, Rectangle)
                Rectangle = New Rectangle(((MarginX * 3) + (ColumnWidth * 2)), (((MarginY * 2) + RowHeight) + MarginTop), ColumnWidth, RowHeight)
                PositionButton(6, Rectangle)
                Rectangle = New Rectangle(MarginX, (((MarginY * 3) + (RowHeight * 2)) + MarginTop), ColumnWidth, RowHeight)
                PositionButton(7, Rectangle)
                Rectangle = New Rectangle(((MarginX * 2) + ColumnWidth), (((MarginY * 3) + (RowHeight * 2)) + MarginTop), ColumnWidth, RowHeight)
                PositionButton(8, Rectangle)
                Rectangle = New Rectangle(((MarginX * 3) + (ColumnWidth * 2)), (((MarginY * 3) + (RowHeight * 2)) + MarginTop), ColumnWidth, RowHeight)
                PositionButton(9, Rectangle)
                Rectangle = New Rectangle(MarginX, (((MarginY * 4) + (RowHeight * 3)) + MarginTop), ColumnWidth, RowHeight)
                PositionButton(10, Rectangle)
                Rectangle = New Rectangle(((MarginX * 2) + ColumnWidth), (((MarginY * 4) + (RowHeight * 3)) + MarginTop), ColumnWidth, RowHeight)
                PositionButton(11, Rectangle)
                Rectangle = New Rectangle(((MarginX * 3) + (ColumnWidth * 2)), (((MarginY * 4) + (RowHeight * 3)) + MarginTop), ColumnWidth, RowHeight)
                PositionButton(12, Rectangle)
            Else
                ColumnWidth = CInt((w - (MarginX * 5)) / 4)
                RowHeight = CInt((h - (MarginY * 4)) / 3)
                Rectangle = New Rectangle(MarginX, (MarginY + MarginTop), ColumnWidth, RowHeight)
                PositionButton(1, Rectangle)
                Rectangle = New Rectangle(((MarginX * 2) + ColumnWidth), (MarginY + MarginTop), ColumnWidth, RowHeight)
                PositionButton(2, Rectangle)
                Rectangle = New Rectangle(((MarginX * 3) + (ColumnWidth * 2)), (MarginY + MarginTop), ColumnWidth, RowHeight)
                PositionButton(3, Rectangle)
                Rectangle = New Rectangle(((MarginX * 4) + (ColumnWidth * 3)), (MarginY + MarginTop), ColumnWidth, RowHeight)
                PositionButton(4, Rectangle)
                Rectangle = New Rectangle(MarginX, (((MarginY * 2) + RowHeight) + MarginTop), ColumnWidth, RowHeight)
                PositionButton(5, Rectangle)
                Rectangle = New Rectangle(((MarginX * 2) + ColumnWidth), (((MarginY * 2) + RowHeight) + MarginTop), ColumnWidth, RowHeight)
                PositionButton(6, Rectangle)
                Rectangle = New Rectangle(((MarginX * 3) + (ColumnWidth * 2)), (((MarginY * 2) + RowHeight) + MarginTop), ColumnWidth, RowHeight)
                PositionButton(7, Rectangle)
                Rectangle = New Rectangle(((MarginX * 4) + (ColumnWidth * 3)), (((MarginY * 2) + RowHeight) + MarginTop), ColumnWidth, RowHeight)
                PositionButton(8, Rectangle)
                Rectangle = New Rectangle(MarginX, (((MarginY * 3) + (RowHeight * 2)) + MarginTop), ColumnWidth, RowHeight)
                PositionButton(9, Rectangle)
                Rectangle = New Rectangle(((MarginX * 2) + ColumnWidth), (((MarginY * 3) + (RowHeight * 2)) + MarginTop), ColumnWidth, RowHeight)
                PositionButton(10, Rectangle)
                Rectangle = New Rectangle(((MarginX * 3) + (ColumnWidth * 2)), (((MarginY * 3) + (RowHeight * 2)) + MarginTop), ColumnWidth, RowHeight)
                PositionButton(11, Rectangle)
                Rectangle = New Rectangle(((MarginX * 4) + (ColumnWidth * 3)), (((MarginY * 3) + (RowHeight * 2)) + MarginTop), ColumnWidth, RowHeight)
                PositionButton(12, Rectangle)
            End If
            IsPrepared = True
        End If
    End Sub

    ''' <summary>
    ''' Positions the button, helper function for autoscale page modes.
    ''' </summary>
    ''' <param name="buttonID">The button_id.</param>
    ''' <param name="geometry">The System.Drawing.Rectangle.</param>
    Private Sub PositionButton(ByVal buttonID As Integer, ByRef geometry As System.Drawing.Rectangle)
        If Buttons.ContainsKey(buttonID) Then
            Dim Btn As Button = Buttons.Item(buttonID)
            Btn.Geometry = geometry
        End If
    End Sub
End Class

