' Copyright 2007 Rubke <interrubke@gmail.com>
' Copyright 2008 - 2009 Sven Strickroth <email@cs-ware.de>
'
' This file is part of OSMtracker.
'
' OSMtracker is free software: you can redistribute it and/or modify
' it under the terms of the GNU General Public License version 3 as
' published by the Free Software Foundation.
'
' OSMtracker is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
' GNU General Public License for more details.
'
' You should have received a copy of the GNU General Public License
' along with OSMtracker. If not, see <http://www.gnu.org/licenses/>.

Imports System.IO
Imports System.Reflection
Module EmbeddedRessourcesSVN
    ''' <summary>
    ''' Gets a io-stream for an embedded resource.
    ''' </summary>
    ''' <param name="resourceName">Name of the resource.</param>
    ''' <returns>a io-stream to the embedded resource</returns>
    Public Function GetResourceStream(ByVal resourceName As String) As System.IO.Stream
        Dim assemblyName As String = System.Reflection.Assembly.GetExecutingAssembly().GetName().Name.ToString()
        Return System.Reflection.Assembly.GetExecutingAssembly().GetManifestResourceStream(assemblyName & "." & resourceName)
    End Function
    ''' <summary>
    ''' Returns the path where the application is stored.
    ''' </summary>
    ''' <returns></returns>
    Public Function GetAppPath() As String
#If Not Win32 Then
        Return Path.GetDirectoryName([Assembly].GetExecutingAssembly().GetModules(0).FullyQualifiedName)
#Else
        Dim m() As [Module]
        m = [Assembly].GetExecutingAssembly().GetModules(False)
        Return Path.GetDirectoryName(m(0).FullyQualifiedName)
#End If
    End Function
    ''' <summary>
    ''' Returns the formatted version.
    ''' </summary>
    ''' <returns></returns>
    Public Function getAppVersion() As String
        Dim Version As System.Version = System.Reflection.Assembly.GetExecutingAssembly().GetName().Version
        Return Version.Major.ToString() & "." & Version.Minor.ToString() & "." & Version.Build.ToString() & "-" & getSVNRevision()
    End Function
    ''' <summary>
    ''' Returns the SVN revision.
    ''' </summary>
    ''' <returns></returns>
    Public Function getSVNRevision() As String
        Dim svnRevisionStream As New IO.StreamReader(GetResourceStream("SVNRevision.txt"))
        Return svnRevisionStream.ReadLine()
    End Function
    ''' <summary>
    ''' Determines whether the compiled application is a development version.
    ''' </summary>
    ''' <returns>
    ''' <c>true</c> if the application is a development version; otherwise, <c>false</c>.
    ''' </returns>
    Public Function isDevelopmentVersion() As Boolean
        Dim Version As System.Version = System.Reflection.Assembly.GetExecutingAssembly().GetName().Version
        Return Not (Version.Minor Mod 2 = 0)
    End Function
End Module
