' Copyright 2008 Sven Strickroth <email@cs-ware.de>
'
' This file is part of OSMtracker.
'
' OSMtracker is free software: you can redistribute it and/or modify
' it under the terms of the GNU General Public License version 3 as
' published by the Free Software Foundation.
'
' OSMtracker is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
' GNU General Public License for more details.
'
' You should have received a copy of the GNU General Public License
' along with OSMtracker. If not, see <http://www.gnu.org/licenses/>.

Option Explicit On
Option Strict On

Namespace GPS
    Public MustInherit Class GPSDevice
        Public MustOverride Function Connect() As Boolean
        Public MustOverride Sub Disconnect()
        Public Event Connected()
        Public Event Disconnected()
        Public Event FixChanged()
        Public Event PositionChanged(ByRef gpsPosition As GPS_POSITION)

        Private WithEvents DataTooOldTimer As Timer

        Protected gpsPosition As New GPS_POSITION

        Private Shared FIX_TYPES() As String = {"N", "F", "D", "F", "4", "5", "EST", "7", "8"}

        Protected lastDataReceivedDate As Date
        Protected metric As Boolean = True
        Private syncTimeRequested As Boolean = False
        Protected gpsConnected As Boolean = False
        Public Property Metrics() As Boolean
            Get
                Return metric
            End Get
            Set(ByVal value As Boolean)
                metric = value
            End Set
        End Property
        Public ReadOnly Property HasFix() As Boolean
            Get
                Return gpsPosition.FixType > 0
            End Get
        End Property
        Public ReadOnly Property FixType() As Integer
            Get
                Return gpsPosition.FixType
            End Get
        End Property
        Public ReadOnly Property FixTypeString() As String
            Get
                Return FIX_TYPES(gpsPosition.FixType)
            End Get
        End Property
        Public ReadOnly Property IsConnected() As Boolean
            Get
                Return gpsConnected
            End Get
        End Property
        Public ReadOnly Property LastDataReceived() As Date
            Get
                Return lastDataReceivedDate
            End Get
        End Property
        Public ReadOnly Property Speed() As Single
            Get
                If Metrics Then
                    Return CSng(gpsPosition.Speed * 1.852)
                Else
                    Return CSng(gpsPosition.Speed * 1.150779)
                End If
            End Get
        End Property
        Public ReadOnly Property Latitude() As Double
            Get
                Return gpsPosition.Latitude
            End Get
        End Property
        Public ReadOnly Property Longitude() As Double
            Get
                Return gpsPosition.Longitude
            End Get
        End Property
        Public ReadOnly Property Altitude() As Integer
            Get
                Return gpsPosition.AltitudeWRTSeaLevel
            End Get
        End Property
        Public ReadOnly Property Bearing() As Single
            Get
                Return gpsPosition.Heading
            End Get
        End Property
        Public ReadOnly Property HDOP() As Double
            Get
                Return gpsPosition.HorizontalDilutionOfPrecision
            End Get
        End Property
        Public ReadOnly Property VDOP() As Double
            Get
                Return gpsPosition.VerticalDilutionOfPrecision
            End Get
        End Property
        Public ReadOnly Property PDOP() As Double
            Get
                Return gpsPosition.PositionDilutionOfPrecision
            End Get
        End Property
        Public ReadOnly Property SatellitesInUse() As Integer
            Get
                Return gpsPosition.SatelliteCount
            End Get
        End Property
        Public Overridable Sub RequestTimeSync()
            syncTimeRequested = True
        End Sub
        <Runtime.InteropServices.DllImport("coredll.dll")> Private Shared Function SetSystemTime(ByRef lpSystemTime As SYSTEMTIME) As Boolean
        End Function
        Protected Sub SyncTime(ByRef position As GPS_POSITION)
            If syncTimeRequested = True Then
                SyncTime(position.GetSYSTEMTIME)
            End If
        End Sub
        Protected Sub SyncTime(ByRef st As SYSTEMTIME)
            If syncTimeRequested = True Then
                syncTimeRequested = False
                If SetSystemTime(st) = True Then
                    MsgBox("time set to: " & System.DateTime.Now.ToString, MsgBoxStyle.Information)
                End If
            End If
        End Sub
        Protected Sub RaiseConnected()
            RaiseEvent Connected()
            'start DataTooOldTimer
            DataTooOldTimer = New Timer
            DataTooOldTimer.Interval = 1500
            DataTooOldTimer.Enabled = True
        End Sub
        Protected Sub RaiseDisconnected()
            'stop DataTooOldTimer
            If DataTooOldTimer IsNot Nothing Then
                DataTooOldTimer.Enabled = False
                DataTooOldTimer = Nothing
            End If
            If gpsPosition.FixType > 0 Then
                gpsPosition.FixType = 0
                RaiseFixChanged()
            Else
                gpsPosition.FixType = 0
            End If
            RaiseEvent Disconnected()
        End Sub
        Protected Sub RaiseFixChanged()
            RaiseEvent FixChanged()
        End Sub
        Protected Sub RaisePositionChanged()
            ' do not leak our instance
            RaiseEvent PositionChanged(DirectCast(gpsPosition.Clone, GPS_POSITION))
        End Sub
        Private Sub DataTooOldTimer_Tick(ByVal sender As Object, ByVal e As System.EventArgs) Handles DataTooOldTimer.Tick
            If Now.Subtract(LastDataReceived).TotalSeconds > 10 Then
                gpsPosition.SatelliteCount = 0
                If HasFix = True Then
                    gpsPosition.FixType = 0
                    RaiseEvent FixChanged()
                End If
            End If
        End Sub
    End Class
End Namespace
