' Copyright 2007 Rubke <interrubke@gmail.com>
' Copyright 2008 - 2009 Sven Strickroth <email@cs-ware.de>
'
' This file is part of OSMtracker.
'
' OSMtracker is free software: you can redistribute it and/or modify
' it under the terms of the GNU General Public License version 3 as
' published by the Free Software Foundation.
'
' OSMtracker is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
' GNU General Public License for more details.
'
' You should have received a copy of the GNU General Public License
' along with OSMtracker. If not, see <http://www.gnu.org/licenses/>.

Option Explicit On
Option Strict On

Public Structure PointD
    Public x As Double
    Public y As Double
    Sub New(ByVal newX As Double, ByVal newY As Double)
        x = newX
        y = newY
    End Sub
    Sub Change(ByVal diffX As Double, ByVal diffY As Double)
        x += diffX
        y += diffY
    End Sub
End Structure

Public Class Projection
    Public Shared Function LatLon2Mercator(ByRef position As PointD) As PointD
        Dim p As New PointD
        p.x = position.x * 20037508.34 / 180
        p.y = (Math.Log(Math.Tan((90 + position.y) * Math.PI / 360)) / (Math.PI / 180)) * 20037508.34 / 180
        Return p
    End Function

    Public Shared Function Mercator2LatLon(ByRef position As PointD) As PointD
        Dim p As New PointD

        p.x = (position.x / 20037508.34) * 180

        p.y = 180 / Math.PI * (2 * Math.Atan(Math.Exp((position.y / 20037508.34) * Math.PI)) - Math.PI / 2)
        Return p
    End Function
End Class
Public Class GPS_POSITION
    Implements ICloneable
    Public UTCTime As DateTime
    Public Latitude As Double
    Public Longitude As Double
    Public Speed As Single
    Public Heading As Single
    Public AltitudeWRTSeaLevel As Integer
    Public FixType As Integer
    Public PositionDilutionOfPrecision As Single
    Public HorizontalDilutionOfPrecision As Single
    Public VerticalDilutionOfPrecision As Single
    Public SatelliteCount As Integer
    Sub Change(ByVal dblLatitude As Double, ByVal dblLongitude As Double)
        Me.Latitude = dblLatitude
        Me.Longitude = dblLongitude
    End Sub
    Public Shared Operator =(ByVal pos1 As GPS_POSITION, ByVal pos2 As GPS_POSITION) As Boolean
        Return (pos1.Latitude = pos2.Latitude) AndAlso (pos1.Longitude = pos2.Longitude)
    End Operator
    Public Shared Operator <>(ByVal pos1 As GPS_POSITION, ByVal pos2 As GPS_POSITION) As Boolean
        Return (pos1.Latitude <> pos2.Latitude) OrElse (pos1.Longitude <> pos2.Longitude)
    End Operator
    Public Sub SetUTCTime(ByVal year As Integer, ByVal month As Integer, ByVal day As Integer, ByVal hour As Integer, ByVal minute As Integer, ByVal second As Integer)
        UTCTime = New DateTime(year, month, day, hour, minute, second, DateTimeKind.Utc)
    End Sub
    Public Sub SetUTCTime(ByRef structre As SYSTEMTIME)
        SetUTCTime(structre.wYear, structre.wMonth, structre.wDay, structre.wHour, structre.wMinute, structre.wSecond)
    End Sub
    Public Function GetSYSTEMTIME() As SYSTEMTIME
        Dim systemtime As SYSTEMTIME
        systemtime.wYear = CUShort(UTCTime.Year)
        systemtime.wMonth = CUShort(UTCTime.Month)
        systemtime.wDay = CUShort(UTCTime.Day)
        systemtime.wHour = CUShort(UTCTime.Hour)
        systemtime.wMinute = CUShort(UTCTime.Minute)
        systemtime.wSecond = CUShort(UTCTime.Second)
        Return systemtime
    End Function
    Public Overridable Function Clone() As Object Implements ICloneable.Clone
        Dim position As New GPS_POSITION
        position.UTCTime = (UTCTime)
        position.Latitude = Latitude
        position.Longitude = Longitude
        position.Speed = Speed
        position.Heading = Heading
        position.AltitudeWRTSeaLevel = AltitudeWRTSeaLevel
        position.FixType = FixType
        position.PositionDilutionOfPrecision = PositionDilutionOfPrecision
        position.HorizontalDilutionOfPrecision = HorizontalDilutionOfPrecision
        position.VerticalDilutionOfPrecision = VerticalDilutionOfPrecision
        position.SatelliteCount = SatelliteCount
        Return position
    End Function
End Class
Public Structure SYSTEMTIME
    <CLSCompliant(False)> _
    Public wYear As UInt16
    <CLSCompliant(False)> _
    Public wMonth As UInt16
    <CLSCompliant(False)> _
    Public wDayOfWeek As UInt16
    <CLSCompliant(False)> _
    Public wDay As UInt16
    <CLSCompliant(False)> _
    Public wHour As UInt16
    <CLSCompliant(False)> _
    Public wMinute As UInt16
    <CLSCompliant(False)> _
    Public wSecond As UInt16
    <CLSCompliant(False)> _
    Public wMilliseconds As UInt16
    Public Shared Operator =(ByVal st1 As SYSTEMTIME, ByVal st2 As SYSTEMTIME) As Boolean
        Return (st1.wSecond = st2.wSecond) AndAlso (st1.wMinute = st2.wMinute) AndAlso (st1.wHour = st2.wHour) AndAlso (st1.wDay = st2.wDay) AndAlso (st1.wMonth = st2.wMonth) AndAlso (st1.wYear = st2.wYear)
    End Operator
    Public Shared Operator <>(ByVal st1 As SYSTEMTIME, ByVal st2 As SYSTEMTIME) As Boolean
        Return Not (st1 = st2)
    End Operator
End Structure
Public Structure MAP_POSITION
    ' "normalize" lat and lon
    Private Const MUL_FRAC As Integer = 10000000
    Private latitude As Single
    Private longitude As Single
    Sub New(ByVal slat As Double, ByVal slon As Double)
        latitude = CSng(CInt(Fix(slat * MUL_FRAC)) / MUL_FRAC)
        longitude = CSng(CInt(Fix(slon * MUL_FRAC)) / MUL_FRAC)
    End Sub
    Sub New(ByRef gpsPosition As GPS_POSITION)
        latitude = CSng(CInt(Fix(gpsPosition.Latitude * MUL_FRAC)) / MUL_FRAC)
        longitude = CSng(CInt(Fix(gpsPosition.Longitude * MUL_FRAC)) / MUL_FRAC)
    End Sub
    Public ReadOnly Property LatF() As Single
        Get
            LatF = latitude
        End Get
    End Property
    Public ReadOnly Property LonF() As Single
        Get
            LonF = longitude
        End Get
    End Property
    Sub Change(ByRef gpsPosition As GPS_POSITION)
        latitude = CSng(CInt(Fix(gpsPosition.Latitude * MUL_FRAC)) / MUL_FRAC)
        longitude = CSng(CInt(Fix(gpsPosition.Longitude * MUL_FRAC)) / MUL_FRAC)
    End Sub
    Sub Change(ByVal slat As Double, ByVal slon As Double)
        Try
            latitude = CSng(CInt(Fix(slat * MUL_FRAC)) / MUL_FRAC)
            longitude = CSng(CInt(Fix(slon * MUL_FRAC)) / MUL_FRAC)
        Catch e As Exception
            ' ignored
        End Try
        If latitude > 89 Then
            latitude = 89
        ElseIf latitude < -89 Then
            latitude = -89
        End If
        If longitude > 180 Then
            longitude = 180
        ElseIf longitude < -180 Then
            longitude = -180
        End If
    End Sub
    Sub ChangeDelta(ByVal lat_delta As Double, ByVal lon_delta As Double)
        Try
            latitude = CSng(CInt(Fix((latitude + lat_delta) * MUL_FRAC)) / MUL_FRAC)
            longitude = CSng(CInt(Fix((longitude + lon_delta) * MUL_FRAC)) / MUL_FRAC)
        Catch e As Exception
            ' ignored
        End Try
        If latitude > 89 Then
            latitude = 89
        ElseIf latitude < -89 Then
            latitude = -89
        End If
        If longitude > 180 Then
            longitude = 180
        ElseIf longitude < -180 Then
            longitude = -180
        End If
    End Sub
    Public Shared Operator =(ByVal pos1 As MAP_POSITION, ByVal pos2 As MAP_POSITION) As Boolean
        Return (pos1.latitude = pos2.latitude) AndAlso (pos1.longitude = pos2.longitude)
    End Operator
    Public Shared Operator <>(ByVal pos1 As MAP_POSITION, ByVal pos2 As MAP_POSITION) As Boolean
        Return (pos1.latitude <> pos2.latitude) OrElse (pos1.longitude <> pos2.longitude)
    End Operator
End Structure