' Copyright 2008 <carsten967@freenet.de>
'
' based on GPXFile:
' Copyright 2008 Sven Strickroth <email@cs-ware.de>
'
' This file is part of OSMtracker.
'
' OSMtracker is free software: you can redistribute it and/or modify
' it under the terms of the GNU General Public License version 3 as
' published by the Free Software Foundation.
'
' OSMtracker is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
' GNU General Public License for more details.
'
' You should have received a copy of the GNU General Public License
' along with OSMtracker. If not, see <http://www.gnu.org/licenses/>.

Option Explicit On
Option Strict On

Imports System.IO

Public Class OSMFile
    Public Shared myCultureInfo As New System.Globalization.CultureInfo("en-US")
    ' recheck these structs
    Public Structure osmPos
        Dim latF As Single
        Dim lonF As Single
    End Structure

    Public Structure osmNode
        Dim name As String
        Dim position As osmPos
        Dim attributes() As String ' 0: type (e.g. shop=supermarket), 1... =opening hours etc.
        Dim iconFile As String ' filename of icon, better refer as inteegr to preloaded iconlist
    End Structure

    Public Structure osmWay
        Dim name As String
        Dim attributes() As String ' 0: type (e.g. shop=supermarket), 1... =opening hours etc.
        Dim nodeList() As osmPos
    End Structure

    Private NodeList() As osmNode
    Private pathName As String ' for osm files
    Private iconPath As String ' for icons
    Private osmfile_mutex As New Threading.Mutex
    Public Sub New(ByVal osmPathName As String, ByVal iconPathName As String)
        pathName = osmPathName
        iconPath = iconPathName
        If File.Exists(pathName & "user-poi.osm") Then
            load("user-poi")
        End If
        If File.Exists(pathName & "region.osm") Then
            load("region")
        End If
    End Sub
    Private Function IconNameResolve(ByVal pAttr0 As String, ByVal defName As String) As String
        ' icon source: http://trac.openstreetmap.org/browser/applications/share/map-icons
        ' default: classic.small
        ' icon parsing better by xml file (e.g. josm template http://josm.openstreetmap.de/browser/trunk/styles/standard/elemstyles.xml )
        Dim retVal As String

        retVal = defName
        If pAttr0.StartsWith("tracker=") Then
            pAttr0 = Replace(pAttr0, "tracker=", "").ToLower()

            If pAttr0 = "atm" Then
                retVal = "money" & IO.Path.DirectorySeparatorChar & "atm.png"
            ElseIf pAttr0 = "bank" Then
                retVal = "money" & IO.Path.DirectorySeparatorChar & "bank.png"
            ElseIf pAttr0 = "taxi" Then
                retVal = "transport" & IO.Path.DirectorySeparatorChar & "taxi.png"
            ElseIf pAttr0 = "bus" Then
                retVal = "transport" & IO.Path.DirectorySeparatorChar & "bus.png"
            ElseIf pAttr0 = "fuel" Then
                retVal = "vehicle" & IO.Path.DirectorySeparatorChar & "fuel_station.png"
            ElseIf pAttr0 = "hotel" Then
                retVal = "accommodation" & IO.Path.DirectorySeparatorChar & "hotel.png"
            ElseIf pAttr0 = "restaurant" Then
                retVal = "food" & IO.Path.DirectorySeparatorChar & "restaurant.png"
            ElseIf pAttr0 = "fast food" Then
                retVal = "food" & IO.Path.DirectorySeparatorChar & "fastfood.png"
            ElseIf pAttr0 = "shop" Then
                retVal = "shopping.png"
            ElseIf pAttr0 = "supermarket" Then
                retVal = "shopping.png"
            ElseIf pAttr0 = "bakery" Then
                retVal = "shopping" & IO.Path.DirectorySeparatorChar & "groceries" & IO.Path.DirectorySeparatorChar & "bakery.png"
            ElseIf pAttr0 = "butcher" Then
                retVal = "shopping" & IO.Path.DirectorySeparatorChar & "groceries" & IO.Path.DirectorySeparatorChar & "butcher.png"
            ElseIf pAttr0 = "sport" Then
                retVal = "sport.png"
                'ElseIf pAttr0 = "worship" Then 
                'retVal = "amenity=place_ofworship"
            ElseIf pAttr0 = "pharmacy" Then
                retVal = "health" & IO.Path.DirectorySeparatorChar & "pharmacy.png"
            ElseIf pAttr0 = "viewpoint" Then
                retVal = "sightseeing" & IO.Path.DirectorySeparatorChar & "viewpoint.png"
            ElseIf pAttr0 = "informat." Then
                retVal = "misc" & IO.Path.DirectorySeparatorChar & "information.png"
            ElseIf pAttr0.StartsWith("max") Then
                retVal = "vehicle" & IO.Path.DirectorySeparatorChar & "restrictions.png"
            ElseIf retVal = "" Then
                If File.Exists(iconPath & "tracker" & IO.Path.DirectorySeparatorChar & pAttr0 & ".png") = True Then
                    retVal = "tracker" & IO.Path.DirectorySeparatorChar & pAttr0 & ".png"
                End If
            End If
        Else
            If pAttr0 = "amenity=atm" Then retVal = "money" & IO.Path.DirectorySeparatorChar & "atm.png"
            If pAttr0 = "amenity=bank" Then retVal = "money" & IO.Path.DirectorySeparatorChar & "bank.png"

            If pAttr0 = "amenity=restaurant" Then retVal = "food" & IO.Path.DirectorySeparatorChar & "restaurant.png"
            If pAttr0 = "amenity=fast_food" Then retVal = "food" & IO.Path.DirectorySeparatorChar & "fastfood.png"

            If pAttr0 = "amenity=taxi" Then retVal = "transport" & IO.Path.DirectorySeparatorChar & "taxi.png"
            If pAttr0 = "amenity=bus_station" Then retVal = "transport" & IO.Path.DirectorySeparatorChar & "bus.png"
            If pAttr0 = "highway=fuel" Then retVal = "vehicle" & IO.Path.DirectorySeparatorChar & "fuel_station.png"

            If pAttr0 = "tourism=information" Then retVal = "misc" & IO.Path.DirectorySeparatorChar & "information.png"
            If pAttr0 = "tourism=viewpoint" Then retVal = "sightseeing" & IO.Path.DirectorySeparatorChar & "viewpoint.png"
            If pAttr0 = "tourism=hotel" Then retVal = "accommodation" & IO.Path.DirectorySeparatorChar & "hotel.png"

            If pAttr0 = "shop=" Then retVal = "shopping.png"
            If pAttr0 = "shop=supermarket" Then retVal = "shopping.png"
            If pAttr0 = "shop=bakery" Then retVal = "shopping" & IO.Path.DirectorySeparatorChar & "groceries" & IO.Path.DirectorySeparatorChar & "bakery.png"
            If pAttr0 = "shop=butcher" Then retVal = "shopping" & IO.Path.DirectorySeparatorChar & "groceries" & IO.Path.DirectorySeparatorChar & "butcher.png"

            If pAttr0 = "sport=multi" Then retVal = "sport.png"

            If pAttr0.StartsWith("maxspeed=") Then retVal = "vehicle" & IO.Path.DirectorySeparatorChar & "restrictions.png"
        End If
        IconNameResolve = retVal
    End Function
    Private Sub parseNodeKey(ByVal idxNode As Integer, ByVal strKey As String, ByVal strValue As String)
        ' only default icons hardcoded

        With NodeList(idxNode)
            ' only selected info held in memory
            If strKey = "name" Then
                .name = strValue
            ElseIf strKey = "place" Then
                .attributes(0) = strKey & "=" & strValue
                .iconFile = IconNameResolve(.attributes(0), "places.png")
            ElseIf strKey = "shop" Then
                .attributes(0) = strKey & "=" & strValue
                .iconFile = IconNameResolve(.attributes(0), "shopping.png")
            ElseIf strKey = "tourism" Then
                .attributes(0) = strKey & "=" & strValue
                .iconFile = IconNameResolve(.attributes(0), "sightseeing.png")
            ElseIf strKey = "historic" Then
                .attributes(0) = strKey & "=" & strValue
                .iconFile = IconNameResolve(.attributes(0), "sightseeing.png")
            ElseIf strKey = "highway" Then
                .attributes(0) = strKey & "=" & strValue
                .iconFile = IconNameResolve(.attributes(0), "transport.png")
            ElseIf strKey = "railway" Then
                .attributes(0) = strKey & "=" & strValue
                .iconFile = IconNameResolve(.attributes(0), "transport.png")
            ElseIf strKey = "sport" Then
                .attributes(0) = strKey & "=" & strValue
                .iconFile = IconNameResolve(.attributes(0), "sport.png")
            ElseIf strKey = "amenity" Then
                .attributes(0) = strKey & "=" & strValue
                .iconFile = IconNameResolve(.attributes(0), "misc.png")
            ElseIf strKey = "opening_hours" Then
                ReDim Preserve .attributes(.attributes.GetLength(0))
                .attributes(.attributes.GetLength(0) - 1) = strKey & "=" & strValue
            End If
        End With
    End Sub
    Private Function str2single(ByVal fStr As String) As Single
        Dim rVal As Single

        fStr.Replace(",", ".")
        rVal = Convert.ToSingle(fStr, myCultureInfo)
        str2single = rVal
    End Function
    Private Sub parseOsmNode(ByVal strContents() As String)
        Dim idxT As String ' N(ode W(ay
        Dim posFnd1 As Integer
        Dim posFnd2 As Integer
        Dim strKey As String
        Dim strValue As String
        Dim saveFlg As Boolean

        idxT = ""
        For Each strLine As String In strContents
            posFnd1 = InStr(strLine, "<node id=")
            If posFnd1 <> 0 Then
                If NodeList Is Nothing Then
                    ReDim NodeList(0)
                Else
                    ReDim Preserve NodeList(NodeList.GetLength(0))
                End If
                idxT = "N"
                With NodeList(NodeList.GetLength(0) - 1)
                    ' init new point
                    .name = ""
                    ReDim .attributes(0)
                    .attributes(0) = ""

                    posFnd1 = InStr(posFnd1, strLine, "lat='")
                    posFnd1 = posFnd1 + 5
                    posFnd2 = InStr(posFnd1, strLine, "'")
                    .position.latF = str2single(Mid(strLine, posFnd1, posFnd2 - posFnd1))
                    posFnd1 = posFnd2 + 1

                    posFnd1 = InStr(posFnd1, strLine, "lon='")
                    posFnd1 = posFnd1 + 5
                    posFnd2 = InStr(posFnd1, strLine, "'")
                    .position.lonF = str2single(Mid(strLine, posFnd1, posFnd2 - posFnd1))
                    posFnd1 = posFnd2 + 1
                End With
            End If

            posFnd1 = InStr(strLine, "</node>")
            If (posFnd1 = 0) And (InStr(strLine, "<node id=") <> 0) Then posFnd1 = InStr(strLine, "/>")
            If posFnd1 <> 0 Then
                ' finish point
                saveFlg = False
                With NodeList(NodeList.GetLength(0) - 1)
                    If .name <> "" Then saveFlg = True
                    If .attributes(0) <> "" Then saveFlg = True
                End With
                If Not saveFlg Then
                    If NodeList.GetLength(0) > 1 Then
                        ReDim Preserve NodeList(NodeList.GetLength(0) - 2)
                    Else
                        NodeList = Nothing
                    End If
                End If
                idxT = " "
            End If

            If idxT = "N" Then
                ' extra point info
                posFnd1 = InStr(strLine, "<tag k='")
                If posFnd1 <> 0 Then
                    posFnd1 = posFnd1 + 8
                    posFnd2 = InStr(posFnd1, strLine, "'")
                    If posFnd2 = 0 Then posFnd2 = InStr(posFnd1, strLine, Chr(34))
                    strKey = Mid(strLine, posFnd1, posFnd2 - posFnd1)
                    If posFnd2 = 0 Then posFnd2 = InStr(posFnd1, strLine, Chr(34))
                    posFnd1 = InStr(posFnd2 + 1, strLine, "v=")
                    posFnd1 = posFnd1 + 3
                    posFnd2 = InStr(posFnd1, strLine, "'")
                    If posFnd2 = 0 Then posFnd2 = InStr(posFnd1, strLine, Chr(34))
                    strValue = Mid(strLine, posFnd1, posFnd2 - posFnd1)
                    Call parseNodeKey(NodeList.GetLength(0) - 1, strKey, strValue)
                End If
            End If
        Next
    End Sub
    Public Sub load(ByVal fileName As String)
        ' load osm file in memory
        Dim strLine As String
        Dim strContents() As String
        Dim posFnd As Integer

        Try
            Dim stream As StreamReader = New StreamReader(New FileStream(pathName & fileName & ".osm", IO.FileMode.Open))
            osmfile_mutex.WaitOne()
            While Not stream.EndOfStream
                strLine = stream.ReadLine

                posFnd = InStr(strLine, "<node id=") ' check for start of node
                If posFnd > 0 Then
                    ReDim strContents(0)
                    strContents(0) = strLine
                    Do
                        posFnd = InStr(strLine, "</node>") ' check for node end
                        If (posFnd = 0) And (InStr(strLine, "<node id=") <> 0) Then posFnd = InStr(strLine, "/>")
                        If posFnd = 0 Then ' next line required
                            strLine = stream.ReadLine
                            ReDim Preserve strContents(strContents.GetLength(0))
                            strContents(strContents.GetLength(0) - 1) = strLine
                        End If
                    Loop Until posFnd > 0 ' "/" found is "end of node"
                    Call parseOsmNode(strContents) ' save node in memory
                End If
            End While
            osmfile_mutex.ReleaseMutex()
            stream.Close()
        Catch ex As Exception
            MsgBox("osm file load error:" & vbCrLf & pathName & fileName & ".osm" & vbCrLf & ex.Message)
        End Try
    End Sub
    Public Function getNodesCount() As Integer
        getNodesCount = NodeList.GetLength(0)
    End Function
    ' Calculate relative pixel coordinates of the position myPos from tile number tile_x x tile_y
    ' HACK should be public in geocalc 
    Private tile_x As Integer, tile_y As Integer, tile_z As Integer
    Const tileWidth As Integer = 256, tileHeight As Integer = 256
    Private powerof2 As Integer() = {1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096, 8192, 16384, 32768, 65536, 131072, 262144, 524288, 1048576}
    Private Sub getXYPosition(ByVal myPos As osmPos, ByVal x As Integer, ByVal y As Integer, ByRef xpos As Integer, ByRef ypos As Integer)
        Dim xf As Double = (myPos.lonF + 180.0) / 360.0 * powerof2(tile_z)
        Dim yf As Double = (1 - Math.Log(Math.Tan(myPos.latF * Math.PI / 180.0) + 1 / Math.Cos(myPos.latF * Math.PI / 180.0)) / Math.PI) / 2.0 * powerof2(tile_z)

        ' get the difference
        xf = xf - x
        yf = yf - y

        ' translate to tile position
        xpos = CInt(xf * tileWidth)
        ypos = CInt(yf * tileHeight)
    End Sub
    Private Function strEscRepl(ByVal inText As String) As String
        ' replace escape sequences
        Dim sPos As Integer
        Dim retText As String
        Dim loopReady As Boolean

        retText = inText
        Do
            loopReady = True

            sPos = retText.IndexOf("&apos;")
            If sPos >= 0 Then
                retText = retText.Substring(0, sPos) & "'" & retText.Substring(sPos + 6)
                loopReady = False
            End If

            sPos = retText.IndexOf("&amp;")
            If sPos >= 0 Then
                retText = retText.Substring(0, sPos) & "&" & retText.Substring(sPos + 6)
                loopReady = False
            End If

        Loop Until loopReady = True
        strEscRepl = retText
    End Function
    Private Sub mapIconDraw(ByVal g As Graphics, ByVal iconName As String, ByVal posX As Integer, ByVal posY As Integer, ByVal pName As String, ByVal pAttr1 As String)
        Dim bmIcon As Bitmap
        Dim drawPen As New Pen(Color.DarkGray, 1)
        Dim drawBrush As New SolidBrush(Color.Blue)
        Dim drawFont As New Font("Arial Unicode MS", 8, FontStyle.Regular)
        Dim drawPos As Point
        Dim dName As String
        Dim tPosY As Single

        ' icon known, load file
        If iconName <> "" Then
            Try
                If File.Exists(iconPath & iconName) = False Then
                    bmIcon = New Bitmap(iconPath & iconName)
                    drawPos.X = posX - Convert.ToInt32(bmIcon.Width / 2)
                    drawPos.Y = posY - Convert.ToInt32(bmIcon.Height / 2)
                    g.DrawImage(bmIcon, drawPos.X, drawPos.Y)
                Else
                    iconName = ""
                End If
            Catch ex As Exception
                iconName = ""
            End Try
        End If

        ' no icon or load failure
        If iconName = "" Then
            g.DrawRectangle(drawPen, posX - 2, posY - 2, 5, 5)
        End If
        If pName <> "" Then
            dName = strEscRepl(pName)
            tPosY = posY - 4
            g.DrawString(dName, drawFont, drawBrush, posX + 5, tPosY)
        End If
        If pAttr1.StartsWith("opening_hours") = True Then
            dName = pAttr1.Substring(14)
            tPosY = tPosY + 10
            g.DrawString(dName, drawFont, drawBrush, posX + 5, tPosY)
        End If
    End Sub
    Public Sub drawIcon(ByRef g As Graphics, ByRef iconName As String, ByVal posX As Integer, ByVal posY As Integer, ByVal pName As String, ByVal pType As String)
        ' draw icon by name 
        ' if name="" (empty string), parse name and set name according to pType
        If iconName = "" Then
            ' waypoint has different type naming, e.g. simply "ATM" instead amenity=atm, so type starts with "tracker=" as a unique key
            iconName = IconNameResolve("tracker=" & pType, "empty.png")
        End If
        Call mapIconDraw(g, iconName, posX, posY, pName, "")
    End Sub
    Public Sub drawOsmLayer(ByRef g As Graphics, ByVal tileX As Integer, ByVal tileY As Integer, ByVal tileZoom As Integer, ByVal tileImageWidth As Integer, ByVal tileImageHeight As Integer)
        ' draw osm layer on map bitmap
        Dim nodeIdx As Integer
        Dim x_pos As Integer, y_pos As Integer
        Dim sBrush As Brush = New SolidBrush(Color.Pink)

        tile_x = tileX : tile_y = tileY : tile_z = tileZoom
        If NodeList IsNot Nothing Then
            For nodeIdx = 0 To NodeList.GetLength(0) - 1
                With NodeList(nodeIdx)
                    getXYPosition(.position, tile_x, tile_y, x_pos, y_pos)
                    If x_pos > 0 And y_pos > 0 And x_pos < tileImageWidth And y_pos < tileImageHeight Then
                        If tileZoom < 14 Then
                            g.FillEllipse(sBrush, x_pos, y_pos, 3, 3)
                        Else
                            ' draw icon
                            Call mapIconDraw(g, .iconFile, x_pos, y_pos, .name, "")
                        End If
                    End If
                End With
            Next
        End If
    End Sub
    ' unused so far
    Public Function osmInfo(ByVal mouseX As Integer, ByVal mouseY As Integer) As String
        ' get info string for osm info at mouse position
        Dim nodeIdx As Integer
        Dim attrIdx As Integer
        Dim tile_x As Integer, tile_y As Integer
        Dim x_pos As Integer, y_pos As Integer
        Dim retVal As String

        retVal = ""
        If NodeList IsNot Nothing Then
            For nodeIdx = 0 To NodeList.GetLength(0) - 1
                With NodeList(nodeIdx)
                    getXYPosition(.position, tile_x, tile_y, x_pos, y_pos)
                    If Math.Abs(mouseX - x_pos) < 10 And Math.Abs(mouseY - y_pos) < 10 Then
                        retVal = .name
                        For attrIdx = 0 To .attributes.GetLength(0)
                            retVal = retVal & vbCrLf & .attributes(attrIdx)
                        Next
                    End If
                End With
            Next
        End If
        osmInfo = retVal
    End Function
End Class
