' Copyright 2007 Rubke <interrubke@gmail.com> (DownloadFile method)
' Copyright 2008 - 2009 Sven Strickroth <email@cs-ware.de>
'
' This file is part of OSMtracker.
'
' OSMtracker is free software: you can redistribute it and/or modify
' it under the terms of the GNU General Public License version 3 as
' published by the Free Software Foundation.
'
' OSMtracker is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
' GNU General Public License for more details.
'
' You should have received a copy of the GNU General Public License
' along with OSMtracker. If not, see <http://www.gnu.org/licenses/>.

Option Explicit On
Option Strict On

Imports System.Collections.Generic
Imports System.Net
Imports System.Threading
Imports System.Runtime.InteropServices
'http://msdn.microsoft.com/de-de/library/bb979403.aspx
'http://msdn.microsoft.com/de-de/library/bb979361.aspx
Public Class OSMTileDownloader
    Public Class TileQueueEntry
        Public url As String
        Public fileName As String
        Public tileX As Integer
        Public tileY As Integer
        Public tileZ As Integer = 13
        Public Sub New(ByVal newURL As String, ByVal newFileName As String, ByVal x As Integer, ByVal y As Integer, ByVal z As Integer)
            url = newURL
            fileName = newFileName
            tileX = x
            tileY = y
            tileZ = z
        End Sub
        Public Overrides Function Equals(ByVal otherObject As Object) As Boolean
            If Not otherObject.GetType.Equals(GetType(TileQueueEntry)) Then
                Return False
            Else
                If DirectCast(otherObject, TileQueueEntry).fileName = fileName Then
                    Return True
                Else
                    Return False
                End If
            End If
        End Function
    End Class
    Public Event DownloadedTile(ByVal downloadedTile As TileQueueEntry)
    Private downloaderThread As Thread = Nothing
    Private queue As Queue(Of TileQueueEntry) = New Queue(Of TileQueueEntry)
    Private newDownloadHandle As IntPtr = IntPtr.Zero
    Private stopHandle As IntPtr = IntPtr.Zero
    Public Sub Download(ByVal newQueueEntry As TileQueueEntry)
        If Not queue.Contains(newQueueEntry) Then
            queue.Enqueue(newQueueEntry)
            StartDownload()
        End If
    End Sub
    Private Sub Downloader()
        Dim abort As Boolean = False
        Dim multiHandler As IntPtr = LocalAlloc(LMEM_ZEROINIT, 8)
        Marshal.WriteInt32(multiHandler, 0, stopHandle.ToInt32())
        Marshal.WriteInt32(multiHandler, 4, newDownloadHandle.ToInt32())
        While Not abort
            While queue.Count > 0
                ResetEvent(newDownloadHandle)
                Dim newEntry As TileQueueEntry
                newEntry = queue.Dequeue()
                If Not IO.File.Exists(newEntry.fileName) Then
                    If DownloadFile(newEntry.url, newEntry.fileName) Then
                        RaiseEvent DownloadedTile(newEntry)
                    End If
                End If
            End While
            If WaitForMultipleObjects(2, multiHandler, 0, -1) = 0 Then
                abort = True
            End If
        End While
        LocalFree(multiHandler)
        CloseHandle(stopHandle)
        stopHandle = Nothing
        CloseHandle(newDownloadHandle)
        newDownloadHandle = Nothing
        downloaderThread = Nothing
    End Sub
    Private Function DownloadFile(ByVal url As String, ByVal fileName As String) As Boolean
        Dim request As WebRequest
        Dim response As WebResponse = Nothing
        Dim input As IO.Stream = Nothing
        Dim output As IO.FileStream = Nothing
        Dim count As Integer = 32 * 1024 ' 32k at a time
        Dim buffer(count - 1) As Byte
        Try
            request = WebRequest.Create(url)
            If request IsNot Nothing Then
                response = request.GetResponse()
                If response IsNot Nothing Then
                    input = response.GetResponseStream()
                    If input IsNot Nothing Then
                        output = New IO.FileStream(fileName, IO.FileMode.Create)
                        If output IsNot Nothing Then
                            Do
                                count = input.Read(buffer, 0, count)
                                If count = 0 Then Exit Do ' ready
                                output.Write(buffer, 0, count)
                            Loop
                        End If
                        DownloadFile = True
                    End If
                End If
            End If

        Catch ex As Exception
            ' MsgBox("error in DownloadFile (" & strUrl & ") reason : " & ex.ToString, MsgBoxStyle.Critical)
        End Try
        If input IsNot Nothing Then input.Close()
        If output IsNot Nothing Then output.Close()
        If response IsNot Nothing Then response.Close()
    End Function
    Private Sub StartDownload()
        If downloaderThread Is Nothing Then
            downloaderThread = New Threading.Thread(New Threading.ThreadStart(AddressOf Me.downloader))
            stopHandle = CreateEvent(IntPtr.Zero, 0, 0, Nothing)
            newDownloadHandle = CreateEvent(IntPtr.Zero, 0, 0, Nothing)
            downloaderThread.Start()
        Else
            SetEvent(newDownloadHandle)
        End If
    End Sub
    Public Sub StopDownloader()
        queue.Clear()
        If stopHandle <> IntPtr.Zero Then
            SetEvent(stopHandle)
        End If
    End Sub
    Public Sub ClearQueue()
        queue.Clear()
    End Sub
End Class
