' Microsoft .NET Compact Framework-based PInvoke Library Sample
' Licensed under the Shared Source License for Microsoft .NET Compact Framework-based PInvoke Library Sample
'
' based on http://msdn.microsoft.com/en-us/library/aa446550.aspx
'
' Geoff Schwab
' Excell Data Corporation
'
' Contributors:
'    Jonathan Wells
'    Microsoft Corporation
'
' Adapted for OSMtracker by:
'     Sven Strickroth
'
' Wave Samples:
'    Seth Demse
'    Microsoft Corporation
'
'    Jonathan Wells
'    Microsoft Corporation

Imports System
Imports System.IO
Imports System.Runtime.InteropServices

'/ <summary>
'/ Encapsulates Wave file properties used internally be WaveIn and WaveOut.
'/ </summary>
Public Class Wave

    'Can be used instead of a device id to open a device
    Public Const WAVE_MAPPER As Integer = -1

    'Flag specifying the use of a callback window for sound messages
    Public Const CALLBACK_WINDOW As Integer = &H10000

    'Error information...
    Private Const WAVERR_BASE As Integer = 32
    Private Const MMSYSERR_BASE As Integer = 0

    'Enum equivalent to MMSYSERR_*
    Public Enum MMSYSERR
        NOERROR = 0
        GENERROR = (MMSYSERR_BASE + 1)
        BADDEVICEID = (MMSYSERR_BASE + 2)
        NOTENABLED = (MMSYSERR_BASE + 3)
        ALLOCATED = (MMSYSERR_BASE + 4)
        INVALHANDLE = (MMSYSERR_BASE + 5)
        NODRIVER = (MMSYSERR_BASE + 6)
        NOMEM = (MMSYSERR_BASE + 7)
        NOTSUPPORTED = (MMSYSERR_BASE + 8)
        BADERRNUM = (MMSYSERR_BASE + 9)
        INVALFLAG = (MMSYSERR_BASE + 10)
        INVALPARAM = (MMSYSERR_BASE + 11)
        HANDLEBUSY = (MMSYSERR_BASE + 12)
        INVALIDALIAS = (MMSYSERR_BASE + 13)
        BADDB = (MMSYSERR_BASE + 14)
        KEYNOTFOUND = (MMSYSERR_BASE + 15)
        READERROR = (MMSYSERR_BASE + 16)
        WRITEERROR = (MMSYSERR_BASE + 17)
        DELETEERROR = (MMSYSERR_BASE + 18)
        VALNOTFOUND = (MMSYSERR_BASE + 19)
        NODRIVERCB = (MMSYSERR_BASE + 20)
        LASTERROR = (MMSYSERR_BASE + 20)
    End Enum

    ' Enum equivalent to WAVERR_*
    Private Enum WAVERR
        NONE = 0
        BADFORMAT = WAVERR_BASE + 0
        STILLPLAYING = WAVERR_BASE + 1
        UNPREPARED = WAVERR_BASE + 2
        SYNC = WAVERR_BASE + 3
        LASTERROR = WAVERR_BASE + 3
    End Enum

    ' <summary>
    ' Used by dwFormats in WAVEINCAPS and WAVEOUTCAPS
    ' Invalid format
    ' </summary>
    Public Const WAVE_INVALIDFORMAT As Integer = &H0
    ' <summary>
    ' Used by dwFormats in WAVEINCAPS and WAVEOUTCAPS
    ' 11.025 kHz, Mono,   8-bit
    ' </summary>
    Public Const WAVE_FORMAT_1M08 As Integer = &H1
    ' <summary>
    ' Used by dwFormats in WAVEINCAPS and WAVEOUTCAPS
    ' 11.025 kHz, Stereo, 8-bit
    ' </summary>
    Public Const WAVE_FORMAT_1S08 As Integer = &H2
    ' <summary>
    ' Used by dwFormats in WAVEINCAPS and WAVEOUTCAPS
    ' 11.025 kHz, Mono,   16-bit
    ' </summary>
    Public Const WAVE_FORMAT_1M16 As Integer = &H4
    ' <summary>
    ' Used by dwFormats in WAVEINCAPS and WAVEOUTCAPS
    ' 11.025 kHz, Stereo, 16-bit
    ' </summary>
    Public Const WAVE_FORMAT_1S16 As Integer = &H8
    ' <summary>
    ' Used by dwFormats in WAVEINCAPS and WAVEOUTCAPS
    ' 22.05  kHz, Mono,   8-bit
    ' </summary>
    Public Const WAVE_FORMAT_2M08 As Integer = &H10
    ' <summary>
    ' Used by dwFormats in WAVEINCAPS and WAVEOUTCAPS
    ' 22.05  kHz, Stereo, 8-bit
    ' </summary>
    Public Const WAVE_FORMAT_2S08 As Integer = &H20
    ' <summary>
    ' Used by dwFormats in WAVEINCAPS and WAVEOUTCAPS
    ' 22.05  kHz, Mono,   16-bit
    ' </summary>
    Public Const WAVE_FORMAT_2M16 As Integer = &H40
    ' <summary>
    ' Used by dwFormats in WAVEINCAPS and WAVEOUTCAPS
    ' 22.05  kHz, Stereo, 16-bit
    ' </summary>
    Public Const WAVE_FORMAT_2S16 As Integer = &H80
    ' <summary>
    ' Used by dwFormats in WAVEINCAPS and WAVEOUTCAPS
    ' 44.1   kHz, Mono,   8-bit
    ' </summary>
    Public Const WAVE_FORMAT_4M08 As Integer = &H100
    ' <summary>
    ' Used by dwFormats in WAVEINCAPS and WAVEOUTCAPS
    ' 44.1   kHz, Stereo, 8-bit
    ' </summary>
    Public Const WAVE_FORMAT_4S08 As Integer = &H200
    ' <summary>
    ' Used by dwFormats in WAVEINCAPS and WAVEOUTCAPS
    ' 44.1   kHz, Mono,   16-bit
    ' </summary>
    Public Const WAVE_FORMAT_4M16 As Integer = &H400
    ' <summary>
    ' Used by dwFormats in WAVEINCAPS and WAVEOUTCAPS
    ' 44.1   kHz, Stereo, 16-bit
    ' </summary>
    Public Const WAVE_FORMAT_4S16 As Integer = &H800

    ' <summary>
    ' WAVEFORMATEX defines the format of waveform-audio data. Only format
    ' information common to all waveform-audio data formats is included in
    ' this structure. For formats requiring additional information, this
    ' structure is included as the first member in another structure, along
    ' with the additional information.
    ' </summary>
    Public Class WAVEFORMATEX

        ' Accessors specifying data positions in a .wave file
        ' RIFF header up to 20 bytes in .wav file
        Protected Const WF_OFFSET_FORMATTAG As Integer = 20
        Protected Const WF_OFFSET_CHANNELS As Integer = 22
        Protected Const WF_OFFSET_SAMPLESPERSEC As Integer = 24
        Protected Const WF_OFFSET_AVGBYTESPERSEC As Integer = 28
        Protected Const WF_OFFSET_BLOCKALIGN As Integer = 32
        Protected Const WF_OFFSET_BITSPERSAMPLE As Integer = 34
        ' Offset 2 for wBitsPerSample
        ' + 4 for the subchunk id "data"
        ' + 4 for the subchunk length
        Public Const WF_OFFSET_DATA As Integer = 44

        ' <summary>
        ' Waveform-audio format type. Format tags are registered with Microsoft
        ' Corporation for many compression algorithms. A complete list of
        ' format tags is located in the Mmsystem.h header file. 
        ' </summary>
        Public Property wFormatTag() As Integer
            Get
                Return Convert.ToInt32(m_wFormatTag)
            End Get
            Set(ByVal Value As Integer)
                m_wFormatTag = Convert.ToUInt16(Value)
            End Set
        End Property
        <CLSCompliant(False)> _
        Protected m_wFormatTag As UInt16 = Convert.ToUInt16(0)
        ' <summary>
        ' Number of channels in the waveform-audio data. Monaural data uses one
        ' channel and stereo data uses two channels.
        ' </summary>
        Public Property nChannels() As Integer
            Get
                Return Convert.ToInt32(m_nChannels)
            End Get
            Set(ByVal Value As Integer)
                m_nChannels = Convert.ToUInt16(Value)
            End Set
        End Property
        <CLSCompliant(False)> _
        Protected m_nChannels As UInt16 = Convert.ToUInt16(0)
        ' <summary>
        ' Sample rate, in samples per second (hertz), that each channel should
        ' be played or recorded. If wFormatTag is WAVE_FORMAT_PCM, then common
        ' values for nSamplesPerSec are 8.0 kHz, 11.025 kHz, 22.05 kHz, and
        ' 44.1 kHz. For non-PCM formats, this member must be computed according
        ' to the manufacturer's specification of the format tag.
        ' </summary>
        Public Property nSamplesPerSec() As Integer
            Get
                Return Convert.ToInt32(m_nSamplesPerSec)
            End Get
            Set(ByVal Value As Integer)
                m_nSamplesPerSec = Convert.ToUInt32(Value)
            End Set
        End Property
        <CLSCompliant(False)> _
        Protected m_nSamplesPerSec As UInt32 = Convert.ToUInt32(0)
        ' <summary>
        ' Required average data-transfer rate, in bytes per second, for the format
        ' tag. If wFormatTag is WAVE_FORMAT_PCM, nAvgBytesPerSec should be equal to
        ' the product of nSamplesPerSec and nBlockAlign. For non-PCM formats, this
        ' member must be computed according to the manufacturer's specification of
        ' the format tag. 
        ' Playback and record software can estimate buffer sizes by using the
        ' nAvgBytesPerSec member. 
        ' </summary>
        Public Property nAvgBytesPerSec() As Integer
            Get
                Return Convert.ToInt32(m_nAvgBytesPerSec)
            End Get
            Set(ByVal Value As Integer)
                m_nAvgBytesPerSec = Convert.ToUInt32(Value)
            End Set
        End Property
        <CLSCompliant(False)> _
        Protected m_nAvgBytesPerSec As UInt32 = Convert.ToUInt32(0)
        ' <summary>
        ' Block alignment, in bytes. The block alignment is the minimum atomic unit
        ' of data for the wFormatTag format type. If wFormatTag is WAVE_FORMAT_PCM,
        ' nBlockAlign should be equal to the product of nChannels and wBitsPerSample
        ' divided by 8 (bits per byte). For non-PCM formats, this member must be
        ' computed according to the manufacturer's specification of the format tag. 
        ' Playback and record software must process a multiple of nBlockAlign bytes
        ' of data at a time. Data written and read from a device must always start
        ' at the beginning of a block. For example, it is illegal to start playback
        ' of PCM data in the middle of a sample (that is, on a non-block-aligned
        ' boundary). 
        ' </summary>
        Public Property nBlockAlign() As Integer
            Get
                Return Convert.ToInt32(m_nBlockAlign)
            End Get
            Set(ByVal Value As Integer)
                m_nBlockAlign = Convert.ToUInt16(Value)
            End Set
        End Property
        <CLSCompliant(False)> _
        Protected m_nBlockAlign As UInt16 = Convert.ToUInt16(0)
        ' <summary>
        ' Bits per sample for the wFormatTag format type. If wFormatTag is
        ' WAVE_FORMAT_PCM, then wBitsPerSample should be equal to 8 or 16. For
        ' non-PCM formats, this member must be set according to the manufacturer's
        ' specification of the format tag. Some compression schemes cannot define
        ' a value for wBitsPerSample, so this member can be zero.
        ' </summary>
        Public Property wBitsPerSample() As Integer
            Get
                Return Convert.ToInt32(m_wBitsPerSample)
            End Get
            Set(ByVal Value As Integer)
                m_wBitsPerSample = Convert.ToUInt16(Value)
            End Set
        End Property
        <CLSCompliant(False)> _
        Protected m_wBitsPerSample As UInt16 = Convert.ToUInt16(0)

        ' <summary>
        ' Seeks the provided Stream to the position at which this structure starts.
        ' Namely, the wFormatTag member.
        ' </summary>
        ' <param name="fs"></param>
        Public Sub SeekTo(ByVal fs As Stream)
            fs.Seek(WF_OFFSET_FORMATTAG, SeekOrigin.Begin)
        End Sub

        ' <summary>
        ' Seeks the provided Stream to the position immediately after this
        ' structure.
        ' </summary>
        ' <param name="fs"></param>
        Public Sub Skip(ByVal fs As Stream)
            fs.Seek(WF_OFFSET_DATA, SeekOrigin.Begin)
        End Sub

        ' <summary>
        ' Read in a WAVEFORMATEX from the given BinaryReader.
        ' </summary>
        ' <param name="rdr">BinaryReader accessing a WAVEFORMATEX.</param>
        ' <returns>The size of the data following the structure</returns>
        Public Sub Read(ByVal rdr As BinaryReader)

            m_wFormatTag = rdr.ReadUInt16()
            m_nChannels = rdr.ReadUInt16()
            m_nSamplesPerSec = rdr.ReadUInt32()
            m_nAvgBytesPerSec = rdr.ReadUInt32()
            m_nBlockAlign = rdr.ReadUInt16()
            m_wBitsPerSample = rdr.ReadUInt16()

            ' Unused subchunk Id and size
            Dim dataId As UInt32 = rdr.ReadUInt32()
            Dim dataLength As UInt32 = rdr.ReadUInt32()

        End Sub

        ' <summary>
        ' Write out a WAVEFORMATEX to the given BinaryWriter.
        ' </summary>
        ' <param name="wrtr">BinaryWriter to receive the WAVEFORMATEX
        ' information</param>
        Public Sub Write(ByVal wrtr As BinaryWriter)

            wrtr.Write(Convert.ToUInt16(m_wFormatTag))
            wrtr.Write(Convert.ToUInt16(m_nChannels))
            wrtr.Write(Convert.ToUInt32(m_nSamplesPerSec))
            wrtr.Write(Convert.ToUInt32(m_nAvgBytesPerSec))
            wrtr.Write(Convert.ToUInt16(m_nBlockAlign))
            wrtr.Write(Convert.ToUInt16(m_wBitsPerSample))

        End Sub
    End Class


    ' <summary>
    ' This structure defines the header used to identify a waveform-audio buffer.
    ' </summary>
    Public Class WAVEHDR
        Implements IDisposable
        ' <summary>
        ' Used by dwFlags in WAVEHDR
        ' Set by the device driver to indicate that it is finished with the buffer
        ' and is returning it to the application.
        ' </summary>
        Public Const WHDR_DONE As Integer = &H1
        ' <summary>
        ' Used by dwFlags in WAVEHDR
        ' Set by Windows to indicate that the buffer has been prepared with the
        ' waveInPrepareHeader or waveOutPrepareHeader function.
        ' </summary>
        Public Const WHDR_PREPARED As Integer = &H2
        ' <summary>
        ' Used by dwFlags in WAVEHDR
        ' This buffer is the first buffer in a loop. This flag is used only with
        ' output buffers.
        ' </summary>
        Public Const WHDR_BEGINLOOP As Integer = &H4
        ' <summary>
        ' Used by dwFlags in WAVEHDR
        ' This buffer is the last buffer in a loop. This flag is used only with
        ' output buffers.
        ' </summary>
        Public Const WHDR_ENDLOOP As Integer = &H8
        ' <summary>
        ' Used by dwFlags in WAVEHDR
        ' Set by Windows to indicate that the buffer is queued for playback.
        ' </summary>
        Public Const WHDR_INQUEUE As Integer = &H10

        ' <summary>
        ' Set in WAVEFORMATEX.wFormatTag to specify PCM data.
        ' </summary>
        Public Const WAVE_FORMAT_PCM As Integer = 1

        ' <summary>
        ' Long pointer to the address of the waveform buffer. This buffer must
        ' be block-aligned according to the nBlockAlign member of the
        ' WAVEFORMATEX structure used to open the device.
        ' </summary>
        Public lpData As IntPtr = IntPtr.Zero
        ' <summary>
        ' Specifies the length, in bytes, of the buffer.
        ' </summary>
        Public dwBufferLength As Integer = 0
        ' <summary>
        ' When the header is used in input, this member specifies how much data
        ' is in the buffer.
        ' </summary>
        Public dwBytesRecorded As Integer = 0
        ' <summary>
        ' Specifies user data.
        ' </summary>
        Public dwUser As Integer = 0
        ' <summary>
        ' Specifies information about the buffer.
        ' </summary>
        Public dwFlags As Integer = 0
        ' <summary>
        ' Specifies the number of times to play the loop. This member is used
        ' only with output buffers.
        ' </summary>
        Public dwLoops As Integer = 0
        ' <summary>
        ' Reserved.
        ' </summary>
        Public lpNext As IntPtr = IntPtr.Zero
        ' <summary>
        ' Reserved.
        ' </summary>
        Public reserved As Integer = 0

        ' <summary>
        ' Read a data buffer from the supplied BinaryReader.  This method will
        ' allocate memory for the data buffer it is not already allocated.
        ' </summary>
        ' <param name="rdr">BinaryReader containing data</param>
        ' <param name="readLength">Size, in bytes, to be read</param>
        ' <returns>MMSYSERR.NOERROR if successful</returns>
        Public Function Read(ByVal rdr As BinaryReader, ByVal readLength As Integer, ByVal align As Integer) As MMSYSERR

            Dim bufferLength As Integer = readLength

            If bufferLength Mod align <> 0 Then
                bufferLength += (align - (bufferLength Mod align))
            End If

            dwBufferLength = bufferLength
            Dim data(readLength - 1) As Byte
            rdr.Read(data, 0, data.Length)

            If lpData.ToInt32() = IntPtr.Zero.ToInt32() Then
                lpData = Memory.LocalAlloc(Memory.LMEM_FIXED, bufferLength)
            End If

            If lpData.ToInt32() = IntPtr.Zero.ToInt32() Then
                Return MMSYSERR.NOMEM
            End If

            Marshal.Copy(data, 0, lpData, data.Length)

            Return MMSYSERR.NOERROR

        End Function

        ' <summary>
        ' Write the contents of the recorded buffer to the supplied
        ' BinaryWriter.
        ' </summary>
        ' <param name="wrtr">BinaryWriter used as write target</param>
        ' <returns>MMSYSERR.NOERROR if successful</returns>
        Public Function Write(ByVal wrtr As BinaryWriter) As MMSYSERR

            If lpData.ToInt32() = IntPtr.Zero.ToInt32() Then
                Return Wave.MMSYSERR.NOMEM
            End If

            Dim data(dwBytesRecorded - 1) As Byte
            Marshal.Copy(lpData, data, 0, data.Length)
            wrtr.Write(data)

            Return Wave.MMSYSERR.NOERROR

        End Function

        ' <summary>
        ' Initialize an instance of a WAVEHDR with the specified buffer
        ' size.
        ' </summary>
        ' <param name="bufferLength">Size, in bytes, of buffer</param>
        ' <param name="init">true=clear data to 0</param>
        ' <returns>MMSYSERR.NOERROR if successful</returns>
        Public Function Init(ByVal bufferLength As Integer, ByVal initData As Boolean) As MMSYSERR

            If lpData.ToInt32() <> IntPtr.Zero.ToInt32() AndAlso dwBufferLength < bufferLength Then
                Memory.LocalFree(lpData)
                lpData = IntPtr.Zero
            End If

            If lpData.ToInt32() = IntPtr.Zero.ToInt32() Then
                lpData = Memory.LocalAlloc(Memory.LMEM_FIXED, bufferLength)
            End If

            dwBufferLength = bufferLength

            If lpData.ToInt32() = IntPtr.Zero.ToInt32() Then
                Return MMSYSERR.NOMEM
            End If

            If initData Then
                Dim i As Integer
                For i = 0 To bufferLength - 1
                    Marshal.WriteByte(lpData, i, 0)
                Next
            End If

            Return MMSYSERR.NOERROR

        End Function

        ' <summary>
        ' Frees any memory allocated for the buffer.
        ' </summary>
        Public Sub Dispose() Implements IDisposable.Dispose
            If lpData.ToInt32() <> IntPtr.Zero.ToInt32() Then
                Memory.LocalFree(lpData)
            End If
        End Sub
    End Class
End Class