' Microsoft .NET Compact Framework-based PInvoke Library Sample
' Licensed under the Shared Source License for Microsoft .NET Compact Framework-based PInvoke Library Sample
'
' based on http://msdn.microsoft.com/en-us/library/aa446550.aspx
'
' Geoff Schwab
' Excell Data Corporation
'
' Contributors:
'    Jonathan Wells
'    Microsoft Corporation
'
' Adapted for OSMtracker by:
'     Sven Strickroth
'
' Wave Samples:
'    Seth Demse
'    Microsoft Corporation
'
'    Jonathan Wells
'    Microsoft Corporation

Imports System
Imports System.Runtime.InteropServices
Imports System.IO
Imports Microsoft.WindowsCE.Forms
Imports System.Threading

'/ <summary>
'/ Encapsulates Waveform Audio Interface playback functions and provides a simple
'/ interface for playing audio.
'/ </summary>
Public Class WaveOut
    Implements IDisposable

    ' <summary>
    ' Supplies an inteface for loading and playing a .wav file.
    ' </summary>
    Public Class WaveFile
        Implements IDisposable

        ' <summary>
        ' Hardware interface instance for this wave file.
        ' </summary>
        Protected m_hwo As IntPtr = IntPtr.Zero

        ' <summary>
        ' Instance of WAVEFORMATEX associated with this file.
        ' </summary>
        Protected m_wfmt As Wave.WAVEFORMATEX = Nothing

        ' <summary>
        ' Buffers used to read and play wave data.
        ' </summary>
        Protected m_whdr() As Wave.WAVEHDR = Nothing

        ' <summary>
        ' Size of each buffer used for playback
        ' </summary>
        Protected m_bufferSize As Integer

        ' <summary>
        ' Number of buffers needed for playback
        ' </summary>
        Protected m_numBlocks As Integer

        ' <summary>
        ' Current buffer being read
        ' </summary>
        Protected m_curBlock As Integer

        ' <summary>
        ' BinaryRead used to access the streaming audio file.
        ' </summary>
        Protected m_rdr As BinaryReader = Nothing

        ' <summary>
        ' Determines if file is done playing.
        ' </summary>
        Public ReadOnly Property Done() As Boolean
            Get
                Return Not m_playing
            End Get
        End Property
        Protected m_playing As Boolean = False

        ' <summary>
        ' Length, in milliseconds, of the audio file.
        ' </summary>
        Public ReadOnly Property Milliseconds() As Integer
            Get
                Return CInt(1000 * m_dataLength / m_wfmt.nAvgBytesPerSec)
            End Get
        End Property
        Protected m_dataLength As Integer


        ' <summary>
        ' Create an instance of a wave file.  Allocate the two buffers that will
        ' be used for streaming audio.
        ' </summary>
        Public Sub New()
            m_whdr = New Wave.WAVEHDR(1) {}
        End Sub
        Public Function Play(ByRef strm As Stream, ByVal hwnd As IntPtr, ByVal bufferSize As Integer, ByVal volLeft As Integer, ByVal volRight As Integer) As Wave.MMSYSERR

            If strm Is Nothing Then
                Return Wave.MMSYSERR.GENERROR
            End If

            m_rdr = New BinaryReader(strm)
            If m_rdr Is Nothing Then
                Return Wave.MMSYSERR.GENERROR
            End If

            m_wfmt = New Wave.WAVEFORMATEX
            m_wfmt.SeekTo(strm)

            ' Read in the WAVEFORMATEX structure and attempt to open the
            ' device for playback.
            m_wfmt.Read(m_rdr)

            Dim result As Wave.MMSYSERR = waveOutOpen(m_hwo, 0, m_wfmt, hwnd, 0, Wave.CALLBACK_WINDOW)
            If result <> Wave.MMSYSERR.NOERROR Then
                Return result
            End If

            m_dataLength = CInt(m_rdr.BaseStream.Length - Wave.WAVEFORMATEX.WF_OFFSET_DATA)

            If m_dataLength = 0 Then Return Wave.MMSYSERR.READERROR

            If bufferSize = 0 Or bufferSize > m_dataLength Then
                m_bufferSize = m_dataLength
            Else
                m_bufferSize = bufferSize >> 1
            End If

            If (m_bufferSize Mod m_wfmt.nBlockAlign) <> 0 Then
                m_bufferSize += m_wfmt.nBlockAlign - (m_bufferSize Mod m_wfmt.nBlockAlign)
            End If

            ' Determine the number of buffer reads required to play the entire
            ' file
            m_numBlocks = CInt(Fix(m_dataLength / m_bufferSize))
            If (m_numBlocks * m_bufferSize) < m_dataLength Then
                m_numBlocks += 1
            End If

            m_whdr(0) = New Wave.WAVEHDR
            m_whdr(1) = New Wave.WAVEHDR

            ' Read in the first buffer
            result = ReadBuffer(0)
            If result <> Wave.MMSYSERR.NOERROR Then
                Return result
            End If

            ' If the entire file fits in the buffer then close the file
            If m_numBlocks = 1 Then
                m_rdr.BaseStream.Close()
                m_rdr.Close()
                m_rdr = Nothing
            End If

            SetVolume(volLeft, volRight)

            ' Start playback of the first buffer
            result = waveOutWrite(m_hwo, m_whdr(0), Marshal.SizeOf(m_whdr(0)))
            If result <> Wave.MMSYSERR.NOERROR Then
                Return result
            End If

            m_curBlock = 0

            ' Create the second buffer.  If the audio is being streamed, this will
            ' be the next audio block, otherwise it will be padding
            Dim loadThread As Thread = New Thread(New ThreadStart(AddressOf LoadBuffer))
            loadThread.Start()

            m_playing = True

            Return Wave.MMSYSERR.NOERROR
        End Function

        ' <summary>
        ' Play a wave file.
        ' </summary>
        ' <param name="curDevice">Hardware device to use for playback</param>
        ' <param name="fileName">Name of file to play</param>
        ' <param name="hwnd">Handle to a message window to use for messaging</param>
        ' <param name="bufferSize">Size of streaming buffers, a 0 value specifies
        ' that the buffer should be created big enough to fit the entire file</param>
        ' <param name="volLeft">Left channel volume level</param>
        ' <param name="volRight">Right channel volume level</param>
        ' <returns>MMSYSERR.NOERROR if successful</returns>
        Public Function Play(ByVal curDevice As Integer, ByVal fileName As String, ByVal hwnd As IntPtr, ByVal bufferSize As Integer, ByVal volLeft As Integer, ByVal volRight As Integer) As Wave.MMSYSERR
            If m_playing Then
                Return Wave.MMSYSERR.NOERROR
            End If

            If Not File.Exists(fileName) Then
                Return Wave.MMSYSERR.GENERROR
            End If

            Dim fi As FileInfo = New FileInfo(fileName)
            If (fi.Attributes And FileAttributes.ReadOnly) <> 0 Then
                fi.Attributes = fi.Attributes Xor FileAttributes.ReadOnly
            End If

            Return Play(New FileStream(fileName, FileMode.Open), hwnd, hwnd.ToInt32, volLeft, volRight)
        End Function

        ' <summary>
        ' Read in the specified buffer.
        ' </summary>
        ' <param name="bufIndex">Index of buffer to use for the read</param>
        ' <returns>MMSYSERR.NOERROR if successful</returns>
        Protected Function ReadBuffer(ByVal bufIndex As Integer) As Wave.MMSYSERR
            If m_rdr Is Nothing Then
                Return Wave.MMSYSERR.VALNOTFOUND
            End If

            Dim readLength As Integer = m_bufferSize
            If bufIndex < m_numBlocks Then
                Dim remainingDataLength As Integer = CInt(m_rdr.BaseStream.Length - m_rdr.BaseStream.Position)
                If m_bufferSize > remainingDataLength Then
                    readLength = remainingDataLength
                End If
            End If

            ' Read in the next block of data
            Dim result As Wave.MMSYSERR = m_whdr(bufIndex).Read(m_rdr, readLength, m_wfmt.nBlockAlign)
            If result <> Wave.MMSYSERR.NOERROR Then
                Return result
            End If

            ' If the header is not prepared then prepare it
            If (m_whdr(bufIndex).dwFlags And Wave.WAVEHDR.WHDR_PREPARED) = 0 Then
                Return waveOutPrepareHeader(m_hwo, m_whdr(bufIndex), Marshal.SizeOf(m_whdr(bufIndex)))
            End If

            Return Wave.MMSYSERR.NOERROR

        End Function

        ' <summary>
        ' Create an empty buffer to append to the end of the sound.  This protects
        ' the playback because the system will sometimes continue reading after the
        ' BlockDone method is called.
        ' </summary>
        ' <param name="bufIndex">Index of buffer to be created</param>
        ' <returns>MMSYSERR.NOERROR is successful</returns>
        Protected Function CreateBuffer(ByVal bufIndex As Integer) As Wave.MMSYSERR

            Dim result As Wave.MMSYSERR = m_whdr(bufIndex).Init(m_bufferSize, True)
            If result <> Wave.MMSYSERR.NOERROR Then
                Return result
            End If

            If (m_whdr(bufIndex).dwFlags And Wave.WAVEHDR.WHDR_PREPARED) = 0 Then
                Return waveOutPrepareHeader(m_hwo, m_whdr(bufIndex), Marshal.SizeOf(m_whdr(bufIndex)))
            End If

            Return Wave.MMSYSERR.NOERROR

        End Function

        ' <summary>
        ' Load a buffer.  If there are blocks left to be read, then data will
        ' be read into the buffer that is not being played.  Otherwise, the buffer
        ' will be filled with 0's for padding to detect the end of playback.
        ' </summary>
        Public Sub LoadBuffer()

            Dim curBuffer As Integer = (m_curBlock + 3) Mod 2
            If m_whdr(curBuffer) Is Nothing Then Exit Sub
            SyncLock (m_whdr(curBuffer))
                If m_curBlock = (m_numBlocks - 1) Then
                    CreateBuffer(curBuffer)
                Else
                    If ReadBuffer(curBuffer) <> Wave.MMSYSERR.NOERROR Then
                        Exit Sub
                    End If
                End If

                waveOutWrite(m_hwo, m_whdr(curBuffer), Marshal.SizeOf(m_whdr(curBuffer)))
            End SyncLock

        End Sub

        ' <summary>
        ' Called by the MessageWindow when the buffer currently being played has
        ' finished.  This method starts the loading of the next block on a
        ' separate thread.  If the current block is the last one then playback is
        ' stopped.
        ' </summary>
        Public Sub BlockDone()

            m_curBlock += 1

            If m_curBlock < m_numBlocks Then

                Debug.Assert((m_whdr((m_curBlock + 2) Mod 2).dwFlags And Wave.WAVEHDR.WHDR_INQUEUE) <> 0, _
                 "ERROR: A sound block finished before the subsequent block was written.")

                Dim loadThread As Thread = New Thread(New ThreadStart(AddressOf LoadBuffer))
                loadThread.Start()
            ElseIf m_curBlock = m_numBlocks Then
                StopPlay()
            End If

        End Sub


        ' <summary>
        ' Stop playing the current file and clean up all resources.
        ' </summary>
        Public Sub StopPlay()

            waveOutReset(m_hwo)

            m_playing = False

            If Not (m_rdr Is Nothing) Then
                m_rdr.BaseStream.Close()
                m_rdr.Close()
                m_rdr = Nothing
            End If

            Dim i As Integer
            For i = 0 To 1

                If Not (m_whdr(i) Is Nothing) Then
                    SyncLock m_whdr(i)

                        If m_hwo.ToInt32() <> IntPtr.Zero.ToInt32() Then
                            waveOutUnprepareHeader(m_hwo, m_whdr(i), Marshal.SizeOf(m_whdr(i)))
                        End If

                        m_whdr(i).Dispose()
                        m_whdr(i) = Nothing

                    End SyncLock
                End If
            Next

            If m_hwo.ToInt32() <> IntPtr.Zero.ToInt32() Then
                waveOutClose(m_hwo)
            End If

            m_hwo = IntPtr.Zero
            m_wfmt = Nothing

        End Sub
        ' <summary>
        ' Get the volume of this sound.
        ' </summary>
        ' <param name="volLeft">Left channel volume level</param>
        ' <param name="volRight">Right channel volume level</param>
        ' <returns>MMSYSERR.NOERROR if successful</returns>
        Public Function GetVolume(ByRef volLeft As Integer, ByRef volRight As Integer) As Wave.MMSYSERR

            Dim vol As Integer = 0

            Dim result As Wave.MMSYSERR = waveOutGetVolume(m_hwo, vol)
            If result <> Wave.MMSYSERR.NOERROR Then
                Return result
            End If

            volLeft = vol And &HFFFF
            volRight = vol >> 16

            Return Wave.MMSYSERR.NOERROR

        End Function

        ' <summary>
        ' Sets the volume of this sound.
        ' </summary>
        ' <param name="volLeft">Left channel volume level</param>
        ' <param name="volRight">Right channel volume level</param>
        ' <returns>MMSYSERR.NOERROR if successful</returns>
        Public Function SetVolume(ByVal volLeft As Integer, ByVal volRight As Integer) As Wave.MMSYSERR

            Dim vol As Integer = (volLeft And &HFFFF) Or (volRight << 16)
            Return waveOutSetVolume(m_hwo, vol)

        End Function

        Public Sub Dispose() Implements IDisposable.Dispose
            StopPlay()
        End Sub
    End Class

    ' <summary>
    ' Defines the MessageWindow used to receive messages from the audio
    ' system.
    ' </summary>
    Public Class SoundMessageWindow
        Inherits MessageWindow

        Public Const MM_WOM_OPEN As Integer = &H3BB
        Public Const MM_WOM_CLOSE As Integer = &H3BC
        Public Const MM_WOM_DONE As Integer = &H3BD

        ' Instance of a playback interface
        Protected m_wo As WaveOut = Nothing

        Public Sub New(ByVal wo As WaveOut)
            m_wo = wo
        End Sub

        Protected Overrides Sub WndProc(ByRef msg As Message)

            Select Case msg.Msg
                ' When this message is encountered, a block is
                ' done playing, so notify the WaveOut instance.
            Case MM_WOM_DONE
                    m_wo.BlockDone(msg.WParam)
            End Select

            MyBase.WndProc(msg)
        End Sub
    End Class

    ' <summary>
    ' Maintain an instance of a MessageWindow that handles audio messages.
    ' </summary>
    Protected m_msgWindow As SoundMessageWindow = Nothing

    ' <summary>
    ' An instance of WaveFile used as the source for playing audio.
    ' </summary>
    Protected m_file As WaveFile = Nothing

    ' <summary>
    ' Create an instance of WaveOut.
    ' </summary>
    Public Sub New()
        m_msgWindow = New SoundMessageWindow(Me)
        m_file = New WaveFile
    End Sub
    ' <summary>
    ' Specifies if playback has finished.
    ' </summary>
    ' <returns>true if no playback is in progress.</returns>
    Public Function Done() As Boolean

        If Not (m_file Is Nothing) Then
            Return m_file.Done
        End If

        Return True

    End Function
    ' <summary>
    ' Determines the length of the audio file, in milliseconds.
    ' </summary>
    ' <returns>Millieconds</returns>
    Public Function Milliseconds() As Integer

        If Not (m_file Is Nothing) Then
            Return m_file.Milliseconds
        End If

        Return 0

    End Function
    Public Function Play(ByRef stream As Stream, ByVal bufferSize As Integer, ByVal volLeft As Integer, ByVal volRight As Integer) As Wave.MMSYSERR
        If Not (m_file Is Nothing) Then
            Return m_file.Play(stream, m_msgWindow.Hwnd, bufferSize, volLeft, volRight)
        End If
    End Function

    ' <summary>
    ' Play a file.
    ' </summary>
    ' <param name="fileName">Name of file to play</param>
    ' <param name="bufferSize">Size of playback buffers</param>
    ' <param name="volLeft">Volume of left channel</param>
    ' <param name="volRight">Volume of right channel</param>
    ' <returns>MMSYSERR.NOERROR if successful</returns>
    Public Function Play(ByVal fileName As String, ByVal bufferSize As Integer, ByVal volLeft As Integer, ByVal volRight As Integer) As Wave.MMSYSERR

        If Not (m_file Is Nothing) Then
            Return m_file.Play(0, fileName, m_msgWindow.Hwnd, bufferSize, volLeft, volRight)
        End If

        Return Wave.MMSYSERR.GENERROR

    End Function

    ' <summary>
    ' A block has finished so notify the WaveFile.  In a more complicated example,
    ' this class might maintain an array of WaveFile instances.  In such a case, the
    ' wParam of the message could be passed from the MM_WIM_DATA message.  This
    ' value represents the m_hwi member of the file that caused the message.
    ' The code might look something like:
    '    Dim f As WaveFile
    '    For Each f In  m_files
    '        If f.m_hwi.ToInt32() = wParam.ToInt32() Then
    '            f.BlockDone();
    '            Exit ForEach
    '        End If
    '    Next f
    ' </summary>
    Public Sub BlockDone(ByVal hwo As IntPtr)

        If Not (m_file Is Nothing) Then
            m_file.BlockDone()
        End If

    End Sub

    ' <summary>
    ' Clean up an allocated resources.
    ' </summary>
    Public Sub Dispose() Implements IDisposable.Dispose

        If Not (m_msgWindow Is Nothing) Then
            m_msgWindow.Dispose()
        End If

        If Not (m_file Is Nothing) Then
            m_file.Dispose()
        End If

    End Sub

    ' <summary>
    ' This function opens a specified waveform output device for playback.
    ' </summary>
    ' <param name="phwo">Address filled with a handle identifying the open
    ' waveform-audio output device. Use the handle to identify the device
    ' when calling other waveform-audio output functions. This parameter might
    ' be NULL if the WAVE_FORMAT_QUERY flag is specified for fdwOpen.</param>
    ' <param name="uDeviceID">Identifier of the waveform-audio output device to
    ' open. It can be either a device identifier or a Handle to an open
    ' waveform-audio input device.</param>
    ' <param name="pwfx">Pointer to a WAVEFORMATEX structure that identifies
    ' the format of the waveform-audio data to be sent to the device. You can
    ' free this structure immediately after passing it to waveOutOpen.</param>
    ' <param name="dwCallback">Specifies the address of a fixed callback function,
    ' an event handle, a handle to a window, or the identifier of a thread to be
    ' called during waveform-audio playback to process messages related to the
    ' progress of the playback. If no callback function is required, this value
    ' can be zero.</param>
    ' <param name="dwInstance">Specifies user-instance data passed to the
    ' callback mechanism. This parameter is not used with the window callback
    ' mechanism.</param>
    ' <param name="fdwOpen">Flags for opening the device.</param>
    ' <returns>MMSYSERR</returns>
    <DllImport("coredll.dll")> _
    Private Shared Function waveOutOpen(ByRef phwo As IntPtr, ByVal uDeviceID As Integer, ByVal pwfx As Wave.WAVEFORMATEX, ByVal dwCallback As IntPtr, ByVal dwInstance As Integer, ByVal fdwOpen As Integer) As Wave.MMSYSERR
    End Function

    ' <summary>
    ' This function queries the current volume setting of a waveform output device.
    ' </summary>
    ' <param name="hwo">Handle to an open waveform-audio output device. This
    ' parameter can also be a device identifier.</param>
    ' <param name="pdwVolume">Pointer to a variable to be filled with the current
    ' volume setting. The low-order word of this location contains the left-channel
    ' volume setting, and the high-order word contains the right-channel setting.
    ' A value of 0xFFFF represents full volume, and a value of 0x0000 is silence. 
    ' If a device does not support both left and right volume control, the low-order
    ' word of the specified location contains the mono volume level.
    ' The full 16-bit setting(s) set with the waveOutSetVolume function is returned,
    ' regardless of whether the device supports the full 16 bits of volume-level
    ' control.</param>
    ' <returns>MMSYSERR</returns>
    <DllImport("coredll.dll")> _
    Protected Shared Function waveOutGetVolume(ByVal how As IntPtr, ByRef pdwVolume As Integer) As Wave.MMSYSERR
    End Function


    ' <summary>
    ' This function sets the volume of a waveform output device.
    ' </summary>
    ' <param name="hwo">Handle to an open waveform-audio output device. This
    ' parameter can also be a device identifier.</param>
    ' <param name="dwVolume">Specifies a new volume setting. The low-order word
    ' contains the left-channel volume setting, and the high-order word contains
    ' the right-channel setting. A value of 0xFFFF represents full volume, and a
    ' value of 0x0000 is silence. 
    ' If a device does not support both left and right volume control, the low-
    ' order word of dwVolume specifies the volume level, and the high-order word
    ' is ignored.</param>
    ' <returns>MMSYSERR</returns>
    <DllImport("coredll.dll")> _
    Protected Shared Function waveOutSetVolume(ByVal how As IntPtr, ByVal dwVolume As Integer) As Wave.MMSYSERR
    End Function

    ' <summary>
    ' This function prepares a waveform data block for playback.
    ' </summary>
    ' <param name="hwo">Handle to the waveform-audio output device.</param>
    ' <param name="pwh">Pointer to a WAVEHDR structure that identifies the data
    ' block to be prepared. The buffer's base address must be aligned with the
    ' respect to the sample size.</param>
    ' <param name="cbwh">Size, in bytes, of the WAVEHDR structure.</param>
    ' <returns>MMSYSERR</returns>
    <DllImport("coredll.dll")> _
    Private Shared Function waveOutPrepareHeader(ByVal hwo As IntPtr, ByVal pwh As Wave.WAVEHDR, ByVal cbwh As Integer) As Wave.MMSYSERR
    End Function

    ' <summary>
    ' This function sends a data block to the specified waveform output device.
    ' </summary>
    ' <param name="hwo">Handle to the waveform-audio output device.</param>
    ' <param name="pwh">Pointer to a WAVEHDR structure containing information
    ' about the data block.</param>
    ' <param name="cbwh">Size, in bytes, of the WAVEHDR structure.</param>
    ' <returns>MMSYSERR</returns>
    <DllImport("coredll.dll")> _
    Private Shared Function waveOutWrite(ByVal how As IntPtr, ByVal pwh As Wave.WAVEHDR, ByVal cbwh As Integer) As Wave.MMSYSERR
    End Function

    ' <summary>
    ' This function cleans up the preparation performed by waveOutPrepareHeader.
    ' The function must be called after the device driver is finished with a data
    ' block. You must call this function before freeing the data buffer.
    ' </summary>
    ' <param name="hwo">Handle to the waveform-audio output device.</param>
    ' <param name="pwh">Pointer to a WAVEHDR structure identifying the data block
    ' to be cleaned up.</param>
    ' <param name="cbwh">Size, in bytes, of the WAVEHDR structure.</param>
    ' <returns>MMSYSERR</returns>
    <DllImport("coredll.dll")> _
    Private Shared Function waveOutUnprepareHeader(ByVal hwo As IntPtr, ByVal pwh As Wave.WAVEHDR, ByVal cbwh As Integer) As Wave.MMSYSERR
    End Function

    ' <summary>
    ' This function closes the specified waveform output device.
    ' </summary>
    ' <param name="hwo">Handle to the waveform-audio output device. If the function
    ' succeeds, the handle is no longer valid after this call.</param>
    ' <returns>MMSYSERR</returns>
    <DllImport("coredll.dll")> _
    Protected Shared Function waveOutClose(ByVal hwo As IntPtr) As Wave.MMSYSERR
    End Function

    ' <summary>
    ' This function stops playback on a specified waveform output device and
    ' resets the current position to 0. All pending playback buffers are marked
    ' as done and returned to the application.
    ' </summary>
    ' <param name="hwo">Handle to the waveform-audio output device.</param>
    ' <returns>MMSYSERR</returns>
    <DllImport("coredll.dll")> _
    Protected Shared Function waveOutReset(ByVal hwo As IntPtr) As Wave.MMSYSERR
    End Function
    ' <summary>
    ' This structure contains timing information for different types of
    ' multimedia data.
    '		typedef struct mmtime_tag 
    '		{
    '			UINT wType; 
    '			union 
    '			{
    '				DWORD ms; 
    '				DWORD sample; 
    '				DWORD cb; 
    '				DWORD ticks; 
    '				struct 
    '				{
    '					BYTE hour; 
    '					BYTE min; 
    '					BYTE sec; 
    '					BYTE frame; 
    '					BYTE fps; 
    '					BYTE dummy; 
    '					BYTE pad[2]
    '				} smpte; 
    '				struct 
    '				{
    '					DWORD songptrpos;
    '				} midi; 
    '			} u; 
    '		} MMTIME;
    ' </summary>
    Protected Class MMTIME

        ' <summary>
        ' Time format.
        ' </summary>
        Public wType As Integer = 0
        ' <summary>
        ' Byte count. Used when wType is TIME_BYTES.
        ' </summary>
        Public cb As Integer = 0

        ' Padding because this is actually a union
        Public pad As Integer = 0

    End Class

    ' Used by MMTIME.wType
    Protected Const TIME_MS As Integer = &H1
    Protected Const TIME_SAMPLES As Integer = &H2
    Protected Const TIME_BYTES As Integer = &H4
    Protected Const TIME_TICKS As Integer = &H20

    ' <summary>
    ' This function retrieves the current playback position of the specified
    ' waveform output device.
    ' </summary>
    ' <param name="hwo">Handle to the waveform-audio output device.</param>
    ' <param name="pmmt">Pointer to an MMTIME structure.</param>
    ' <param name="cbmmt">Size, in bytes, of the MMTIME structure.</param>
    ' <returns>MMSYSERR</returns>
    <DllImport("coredll.dll")> _
    Protected Shared Function waveOutGetPosition(ByVal hwo As IntPtr, ByVal pmmt As MMTIME, ByVal cbmmt As Integer) As Wave.MMSYSERR
    End Function

    ' <summary>
    ' This structure describes the capabilities of a waveform-audio output device.
    '		typedef struct 
    '		{
    '			WORD wMid; 
    '			WORD wPid; 
    '			MMVERSION vDriverVersion; 
    '			TCHAR szPname[MAXPNAMELEN]; 
    '			DWORD dwFormats; 
    '			WORD wChannels; 
    '			WORD wReserved1; 
    '			DWORD dwSupport;} 
    '		WAVEOUTCAPS;
    '	This structure has an embedded TCHAR array so the managed implementation is
    '	a byte array with accessors.
    ' </summary>
    Protected Class WAVEOUTCAPS

        Const WAVEOUTCAPS_SIZE As Integer = 84

        Private m_data() As Byte = Nothing

        Public ReadOnly Property Size() As Integer
            Get
                Return WAVEOUTCAPS_SIZE
            End Get
        End Property

        Public ReadOnly Property Data() As Byte()
            Get
                Return m_data
            End Get
        End Property

        ' <summary>
        ' Used by dwSupport in WAVEOUTCAPS
        ' Supports pitch control
        ' </summary>
        Public Const WAVECAPS_PITCH As Integer = &H1
        ' <summary>
        ' Used by dwSupport in WAVEOUTCAPS
        ' Supports playback rate control
        ' </summary>
        Public Const WAVECAPS_PLAYBACKRATE As Integer = &H2
        ' <summary>
        ' Used by dwSupport in WAVEOUTCAPS
        ' Supports volume control
        ' </summary>
        Public Const WAVECAPS_VOLUME As Integer = &H4
        ' <summary>
        ' Used by dwSupport in WAVEOUTCAPS
        ' Supports separate left-right volume control
        ' </summary>
        Public Const WAVECAPS_LRVOLUME As Integer = &H8

        ' <summary>
        ' Manufacturer identifier for the device driver for the device.
        ' Manufacturer identifiers are defined in Manufacturer and Product
        ' Identifiers.
        ' </summary>
        Public ReadOnly Property wMid() As Integer
            Get
                Return Convert.ToInt32(BitConverter.ToUInt16(m_data, 0))
            End Get
        End Property
        ' <summary>
        ' Product identifier for the device. Product identifiers are defined in
        ' Manufacturer and Product Identifiers.
        ' </summary>
        Public ReadOnly Property wPid() As Integer
            Get
                Return Convert.ToInt32(BitConverter.ToUInt16(m_data, 2))
            End Get
        End Property
        ' <summary>
        ' Version number of the device driver for the device. The high-order byte
        ' is the major version number, and the low-order byte is the minor version
        ' number.
        ' </summary>
        Public ReadOnly Property vDriverVersion() As Integer
            Get
                Return Convert.ToInt32(BitConverter.ToUInt32(m_data, 4))
            End Get
        End Property
        ' <summary>
        ' Specifies the standard formats that are supported.
        ' </summary>
        Public ReadOnly Property dwFormats() As Integer
            Get
                Return Convert.ToInt32(BitConverter.ToUInt32(m_data, 72))
            End Get
        End Property
        ' <summary>
        ' Number specifying whether the device supports mono (1) or stereo (2)
        ' output.
        ' </summary>
        Public ReadOnly Property wChannels() As Integer
            Get
                Return Convert.ToInt32(BitConverter.ToUInt16(m_data, 76))
            End Get
        End Property
        ' <summary>
        ' Packing.
        ' </summary>
        Public ReadOnly Property wReserved1() As Integer
            Get
                Return Convert.ToInt32(BitConverter.ToUInt16(m_data, 78))
            End Get
        End Property
        ' <summary>
        ' Specifies the optional functionality supported by the device.
        ' </summary>
        Public ReadOnly Property dwSupport() As Integer
            Get
                Return Convert.ToInt32(BitConverter.ToUInt16(m_data, 80))
            End Get
        End Property
        ' <summary>
        ' Null-terminated string that contains the product name.
        ' </summary>
        Public ReadOnly Property szPname() as string
            Get
                Dim bytes(31) As Char
                Dim i As Integer
                For i = 0 To 31
                    bytes(i) = Convert.ToChar(BitConverter.ToUInt16(m_data, i * 2 + 8))
                Next

                Return New String(bytes)
            End Get
        End Property

        Public Sub New()
            m_data = New Byte(WAVEOUTCAPS_SIZE - 1) {}
        End Sub

    End Class

    ' <summary>
    ' This function queries a specified waveform device to determine its
    ' capabilities.
    ' </summary>
    ' <param name="uDeviceID">Identifier of the waveform-audio output device.
    ' It can be either a device identifier or a Handle to an open waveform-audio
    ' output device.</param>
    ' <param name="pwoc">Pointer to a WAVEOUTCAPS structure to be filled with
    ' information about the capabilities of the device.</param>
    ' <param name="cbwoc">Size, in bytes, of the WAVEOUTCAPS structure.</param>
    ' <returns>MMSYSERR</returns>
    <DllImport("coredll.dll")> _
    Protected Shared Function waveOutGetDevCaps(ByVal uDeviceID As Integer, ByVal pwoc() As Byte, ByVal cbwoc As Integer) As Wave.MMSYSERR
    End Function
End Class 'WaveOut
